# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import SkipTest

from trove.common import utils
from trove.tests.scenario import runners
from trove.tests.scenario.runners.test_runners import SkipKnownBug
from trove.tests.scenario.runners.test_runners import TestRunner
from troveclient.compat import exceptions


class RootActionsRunner(TestRunner):

    def __init__(self):
        self.current_root_creds = None
        self.restored_root_creds = None
        self.restored_root_creds2 = None
        super(RootActionsRunner, self).__init__()

    def run_check_root_never_enabled(self, expected_http_code=200):
        self.assert_root_disabled(self.instance_info.id, expected_http_code)

    def assert_root_disabled(self, instance_id, expected_http_code):
        self._assert_root_state(instance_id, False, expected_http_code,
                                "The root has already been enabled on the "
                                "instance.")

    def _assert_root_state(self, instance_id, expected_state,
                           expected_http_code, message):
        # The call returns a nameless user object with 'rootEnabled' attribute.
        client = self.auth_client
        response = client.root.is_root_enabled(instance_id)
        self.assert_client_code(client, expected_http_code)
        actual_state = getattr(response, 'rootEnabled', None)
        self.assert_equal(expected_state, actual_state, message)

    def run_disable_root_before_enabled(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        self.assert_root_disable_failure(
            self.instance_info.id, expected_exception, expected_http_code)

    def assert_root_disable_failure(self, instance_id, expected_exception,
                                    expected_http_code):
        client = self.auth_client
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.root.delete, instance_id)

    def run_enable_root_no_password(self, expected_http_code=200):
        root_credentials = self.test_helper.get_helper_credentials_root()
        self.current_root_creds = self.assert_root_create(
            self.instance_info.id, None, root_credentials['name'],
            expected_http_code)
        self.restored_root_creds = list(self.current_root_creds)

    def assert_root_create(self, instance_id, root_password,
                           expected_root_name, expected_http_code):
        client = self.auth_client
        if root_password is not None:
            root_creds = client.root.create_instance_root(
                instance_id, root_password)
            self.assert_equal(root_password, root_creds[1])
        else:
            root_creds = client.root.create(instance_id)

        self.assert_client_code(client, expected_http_code)
        if expected_root_name is not None:
            self.assert_equal(expected_root_name, root_creds[0])

        self.assert_can_connect(instance_id, root_creds)

        return root_creds

    def assert_can_connect(self, instance_id, test_connect_creds):
        self._assert_connect(instance_id, True, test_connect_creds)

    def _assert_connect(self, instance_id, expected_response,
                        test_connect_creds):
        host = self.get_instance_host(instance_id=instance_id)
        self.report.log(
            "Pinging instance %s with credentials: %s, database: %s" %
            (instance_id, test_connect_creds,
             self.test_helper.credentials.get("database"))
        )

        ping_response = self.test_helper.ping(
            host,
            username=test_connect_creds[0],
            password=test_connect_creds[1],
            database=self.test_helper.credentials.get("database")
        )

        self.assert_equal(expected_response, ping_response)

    def run_check_root_enabled(self, expected_http_code=200):
        self.assert_root_enabled(self.instance_info.id, expected_http_code)

    def assert_root_enabled(self, instance_id, expected_http_code):
        self._assert_root_state(instance_id, True, expected_http_code,
                                "The root has not been enabled on the "
                                "instance yet.")

    def run_enable_root_with_password(self, expected_http_code=200):
        root_credentials = self.test_helper.get_helper_credentials_root()
        password = root_credentials['password']
        if password is not None:
            self.current_root_creds = self.assert_root_create(
                self.instance_info.id,
                password, root_credentials['name'],
                expected_http_code)
        else:
            raise SkipTest("No valid root password defined in %s."
                           % self.test_helper.get_class_name())

    def run_disable_root(self, expected_http_code=204):
        self.restored_root_creds2 = list(self.current_root_creds)
        self.assert_root_disable(self.instance_info.id, expected_http_code)

    def assert_root_disable(self, instance_id, expected_http_code):
        client = self.auth_client
        client.root.delete(instance_id)
        self.assert_client_code(client, expected_http_code)
        self.assert_cannot_connect(self.instance_info.id,
                                   self.current_root_creds)

    def assert_cannot_connect(self, instance_id, test_connect_creds):
        self._assert_connect(instance_id, False, test_connect_creds)

    def run_check_root_still_enabled_after_disable(
            self, expected_http_code=200):
        self.assert_root_enabled(self.instance_info.id, expected_http_code)

    def run_delete_root(self, expected_exception=exceptions.BadRequest,
                        expected_http_code=400):
        self.assert_root_delete_failure(
            self.instance_info.id, expected_exception, expected_http_code)

    def assert_root_delete_failure(self, instance_id, expected_exception,
                                   expected_http_code):
        root_user_name = self.current_root_creds[0]
        client = self.auth_client
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.users.delete,
                           instance_id, root_user_name)

    def run_check_root_enabled_after_restore(
            self, restored_instance_id, restored_creds,
            expected_http_code=200):
        self.assert_root_enabled_after_restore(
            restored_instance_id, restored_creds, True, expected_http_code)

    def run_check_root_enabled_after_restore2(
            self, restored_instance_id, restored_creds,
            expected_http_code=200):
        self.assert_root_enabled_after_restore(
            restored_instance_id, restored_creds, False, expected_http_code)

    def assert_root_enabled_after_restore(
            self, restored_instance_id, restored_creds,
            expected_connect_response, expected_http_code):
        if restored_instance_id:
            self.assert_root_enabled(restored_instance_id, expected_http_code)
            self._assert_connect(restored_instance_id,
                                 expected_connect_response, restored_creds)
        else:
            raise SkipTest("No restored instance.")

    def check_root_disable_supported(self):
        """Throw SkipTest if root-disable is not supported."""
        pass

    def check_inherit_root_state_supported(self):
        """Throw SkipTest if inherting root state is not supported."""
        pass


class PerconaRootActionsRunner(RootActionsRunner):

    def check_root_disable_supported(self):
        raise SkipTest("Operation is currently not supported.")


class MariadbRootActionsRunner(RootActionsRunner):

    def check_root_disable_supported(self):
        raise SkipTest("Operation is currently not supported.")


class PxcRootActionsRunner(RootActionsRunner):

    def check_root_disable_supported(self):
        raise SkipTest("Operation is currently not supported.")


class PostgresqlRootActionsRunner(RootActionsRunner):

    def check_root_disable_supported(self):
        raise SkipTest("Operation is currently not supported.")

    def run_enable_root_with_password(self):
        raise SkipTest("Operation is currently not supported.")

    def run_delete_root(self):
        raise SkipKnownBug(runners.BUG_WRONG_API_VALIDATION)


class CouchbaseRootActionsRunner(RootActionsRunner):

    def _assert_connect(
            self, instance_id, expected_response, test_connect_creds):
        host = self.get_instance_host(instance_id=instance_id)
        self.report.log("Pinging instance %s with credentials: %s"
                        % (instance_id, test_connect_creds))
        mgmt_port = 8091
        mgmt_creds = '%s:%s' % (test_connect_creds[0], test_connect_creds[1])
        rest_endpoint = ('http://%s:%d/pools/nodes'
                         % (host, mgmt_port))
        out, err = utils.execute_with_timeout(
            'curl', '-u', mgmt_creds, rest_endpoint)
        self.assert_equal(expected_response, out and len(out) > 0)

    def check_root_disable_supported(self):
        raise SkipTest("Operation is currently not supported.")

    def run_enable_root_with_password(self):
        raise SkipTest("Operation is currently not supported.")

    def run_delete_root(self):
        raise SkipKnownBug(runners.BUG_WRONG_API_VALIDATION)


class RedisRootActionsRunner(RootActionsRunner):

    def check_inherit_root_state_supported(self):
        raise SkipTest("Redis instances does not inherit root state "
                       "from backups.")



# Copyright 2014 eBay Software Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import ForeignKey
from sqlalchemy.schema import Column
from sqlalchemy.schema import Index
from sqlalchemy.schema import MetaData

from trove.db.sqlalchemy.migrate_repo.schema import Boolean
from trove.db.sqlalchemy.migrate_repo.schema import create_tables
from trove.db.sqlalchemy.migrate_repo.schema import DateTime
from trove.db.sqlalchemy.migrate_repo.schema import Integer
from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table


meta = MetaData()

clusters = Table(
    'clusters',
    meta,
    Column('id', String(36), primary_key=True, nullable=False),
    Column('created', DateTime(), nullable=False),
    Column('updated', DateTime(), nullable=False),
    Column('name', String(255), nullable=False),
    Column('task_id', Integer(), nullable=False),
    Column('tenant_id', String(36), nullable=False),
    Column("datastore_version_id", String(36),
           ForeignKey('datastore_versions.id'), nullable=False),
    Column('deleted', Boolean()),
    Column('deleted_at', DateTime()),
    Index("clusters_tenant_id", "tenant_id"),
    Index("clusters_deleted", "deleted"),)


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    Table('datastores', meta, autoload=True)
    Table('datastore_versions', meta, autoload=True)
    instances = Table('instances', meta, autoload=True)
    create_tables([clusters])
    instances.create_column(Column('cluster_id', String(36),
                                   ForeignKey("clusters.id")))
    instances.create_column(Column('shard_id', String(36)))
    instances.create_column(Column('type', String(64)))
    cluster_id_idx = Index("instances_cluster_id", instances.c.cluster_id)
    cluster_id_idx.create()



#    Copyright 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from trove.common import cfg
from trove.common import timeutils
from trove.common import utils
from trove.db import models as dbmodels

CONF = cfg.CONF


def enum(**enums):
    return type('Enum', (), enums)


class Quota(dbmodels.DatabaseModelBase):
    """Defines the base model class for a quota."""

    _data_fields = ['created', 'updated', 'tenant_id', 'resource',
                    'hard_limit']
    _table_name = 'quotas'

    def __init__(self, tenant_id, resource, hard_limit,
                 id=utils.generate_uuid(), created=timeutils.utcnow(),
                 update=timeutils.utcnow()):
        self.tenant_id = tenant_id
        self.resource = resource
        self.hard_limit = hard_limit
        self.id = id
        self.created = created
        self.update = update


class QuotaUsage(dbmodels.DatabaseModelBase):
    """Defines the quota usage for a tenant."""

    _data_fields = ['created', 'updated', 'tenant_id',
                    'in_use', 'reserved', 'resource']
    _table_name = 'quota_usages'


class Reservation(dbmodels.DatabaseModelBase):
    """Defines the reservation for a quota."""

    _data_fields = ['created', 'updated', 'usage_id', 'delta', 'status']
    _table_name = 'reservations'

    Statuses = enum(NEW='New',
                    RESERVED='Reserved',
                    COMMITTED='Committed',
                    ROLLEDBACK='Rolled Back')


def persisted_models():
    return {
        'quotas': Quota,
        'quota_usages': QuotaUsage,
        'reservations': Reservation,
    }


class Resource(object):
    """Describe a single resource for quota checking."""

    INSTANCES = 'instances'
    RAM = 'ram'
    VOLUMES = 'volumes'
    BACKUPS = 'backups'

    def __init__(self, name, flag=None):
        """
        Initializes a Resource.

        :param name: The name of the resource, i.e., "volumes".
        :param flag: The name of the flag or configuration option
                     which specifies the default value of the quota
                     for this resource.
        """

        self.name = name
        self.flag = flag

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return (isinstance(other, Resource) and
                self.name == other.name and
                self.flag == other.flag)

    @property
    def default(self):
        """Return the default value of the quota."""

        return CONF[self.flag] if self.flag is not None else -1



# Copyright 2017 Eayun, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from trove.extensions.common.views import UserView


class RedisRootCreatedView(UserView):
    def __init__(self, user, failed_slaves):
        self.failed_slaves = failed_slaves
        super(RedisRootCreatedView, self).__init__(user)

    def data(self):
        user_dict = {
            "name": self.user.name,
            "password": self.user.password
        }
        return {"user": user_dict, "failed_slaves": self.failed_slaves}



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from trove.common import cfg
from trove.instance.views import InstanceDetailView
from trove.instance.views import InstanceView
from trove.tests.unittests import trove_testtools
from unittest.mock import MagicMock
from unittest.mock import Mock

CONF = cfg.CONF


class InstanceViewsTest(trove_testtools.TestCase):

    def setUp(self):
        super(InstanceViewsTest, self).setUp()
        self.addresses = [{
            'type': 'private',
            'address': '123.123.123.123',
            'network': 'net-id-private'}, {
            'type': 'private',
            'address': '10.123.123.123',
            'network': 'net-id-private'}, {
            'type': 'public',
            'address': '15.123.123.123',
            'network': 'net-id-public'}]
        self.orig_ip_regex = CONF.ip_regex

    def tearDown(self):
        super(InstanceViewsTest, self).tearDown()
        CONF.ip_regex = self.orig_ip_regex


class InstanceDetailViewTest(trove_testtools.TestCase):

    def setUp(self):
        super(InstanceDetailViewTest, self).setUp()
        self.build_links_method = InstanceView._build_links
        self.build_config_method = InstanceDetailView._build_configuration_info
        InstanceView._build_links = Mock()
        InstanceDetailView._build_configuration_info = Mock()
        self.instance = Mock()
        self.instance.created = 'Yesterday'
        self.instance.updated = 'Now'
        self.instance.datastore_version = Mock()
        self.instance.datastore_version.name = 'mysql_test_version'
        self.instance.datastore_version.manager = 'mysql'
        self.instance.hostname = 'test.trove.com'
        self.ip = "1.2.3.4"
        self.instance.addresses = [
            {"address": self.ip, 'type': 'private', 'network': 'net-id'}]
        self.instance.volume_used = '3'
        self.instance.root_password = 'iloveyou'
        self.instance.get_visible_ip_addresses.return_value = [
            {'type': 'private', 'address': '1.2.3.4', 'network': 'net-id'}]
        self.instance.slave_of_id = None
        self.instance.slaves = []
        self.instance.locality = 'affinity'
        self.instance.server_id = 'server_abc'
        self.instance.volume_id = 'volume_abc'
        self.fault_message = 'Error'
        self.fault_details = 'details'
        self.fault_date = 'now'
        self.instance.fault = Mock()
        self.instance.fault.message = self.fault_message
        self.instance.fault.details = self.fault_details
        self.instance.fault.updated = self.fault_date
        self.context = trove_testtools.TroveTestContext(self)
        self.req = Mock()
        self.req.environ = Mock()
        self.req.environ.__getitem__ = Mock(return_value=self.context)

    def tearDown(self):
        super(InstanceDetailViewTest, self).tearDown()
        InstanceView._build_links = self.build_links_method
        InstanceDetailView._build_configuration_info = self.build_config_method

    def test_data_hostname(self):
        view = InstanceDetailView(self.instance, self.req)
        result = view.data()
        self.assertEqual(self.instance.created, result['instance']['created'])
        self.assertEqual(self.instance.updated, result['instance']['updated'])
        self.assertEqual(self.instance.datastore_version.name,
                         result['instance']['datastore']['version'])
        self.assertEqual(self.instance.hostname,
                         result['instance']['hostname'])
        self.assertNotIn('ip', result['instance'])

    def test_data_ip(self):
        self.instance.hostname = None
        view = InstanceDetailView(self.instance, self.req)
        result = view.data()
        self.assertEqual(self.instance.created, result['instance']['created'])
        self.assertEqual(self.instance.updated, result['instance']['updated'])
        self.assertEqual(self.instance.datastore_version.name,
                         result['instance']['datastore']['version'])
        self.assertNotIn('hostname', result['instance'])
        self.assertEqual([self.ip], result['instance']['ip'])
        self.assertEqual(self.ip,
                         result['instance']['addresses'][0]['address'])

    def test_locality(self):
        self.instance.hostname = None
        view = InstanceDetailView(self.instance, self.req)
        result = view.data()
        self.assertEqual(self.instance.locality,
                         result['instance']['locality'])

    def test_fault(self):
        view = InstanceDetailView(self.instance, self.req)
        result = view.data()
        self.assertEqual(self.fault_message,
                         result['instance']['fault']['message'])
        self.assertEqual(self.fault_details,
                         result['instance']['fault']['details'])
        self.assertEqual(self.fault_date,
                         result['instance']['fault']['created'])

    def test_admin_view(self):
        self.context.is_admin = True
        view = InstanceDetailView(self.instance, self.req)
        result = view.data()
        self.assertIn('server_id', result['instance'])
        self.assertIn('volume_id', result['instance'])

    def test_non_admin_view(self):
        self.context.is_admin = False
        view = InstanceDetailView(self.instance, self.req)
        result = view.data()
        self.assertNotIn('server_id', result['instance'])
        self.assertNotIn('volume_id', result['instance'])

    def test_access(self):
        instance = MagicMock()
        instance.hostname = None
        instance.get_visible_ip_addresses.return_value = [
            {'address': '10.111.0.27', 'type': 'private', 'network': 'net-id'}
        ]
        instance.access = None
        instance.slaves = []

        view = InstanceDetailView(instance, self.req)
        data = view.data()['instance']

        self.assertFalse(data['access']['is_public'])

        instance.get_visible_ip_addresses.return_value = [
            {'address': '10.111.0.27', 'type': 'private', 'network': 'net-id'},
            {'address': '172.30.5.107', 'type': 'public', 'network': 'net-id'}
        ]

        view = InstanceDetailView(instance, self.req)
        data = view.data()['instance']

        self.assertTrue(data['access']['is_public'])



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc


class Storage(object):
    """Base class for Storage driver implementation."""

    @abc.abstractmethod
    def save(self, stream, metadata=None, **kwargs):
        """Persist information from the stream.

        Should return the new backup checkshum and location.
        """

    @abc.abstractmethod
    def load(self, location, backup_checksum, **kwargs):
        """Load a stream from the data location.

        Should return an object that provides "read" method.
        """

    def load_metadata(self, parent_location, parent_checksum):
        """Load metadata for a parent backup.

        It's up to the storage driver to decide how to implement this function.
        """
        return {}

    def is_incremental_backup(self, location):
        """Check if the location is an incremental backup."""
        return False

    @abc.abstractmethod
    def get_backup_lsn(self, location):
        """Get the backup LSN."""



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import ForeignKey
from sqlalchemy.schema import Column
from sqlalchemy.schema import MetaData

from trove.db.sqlalchemy.migrate_repo.schema import Boolean
from trove.db.sqlalchemy.migrate_repo.schema import create_tables
from trove.db.sqlalchemy.migrate_repo.schema import DateTime
from trove.db.sqlalchemy.migrate_repo.schema import Integer
from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table


meta = MetaData()

security_groups = Table(
    'security_groups',
    meta,
    Column('id', String(length=36), primary_key=True, nullable=False),
    Column('name', String(length=255)),
    Column('description', String(length=255)),
    Column('user', String(length=255)),
    Column('tenant_id', String(length=255)),
    Column('created', DateTime()),
    Column('updated', DateTime()),
    Column('deleted', Boolean(), default=0),
    Column('deleted_at', DateTime()),
)

security_group_instance_associations = Table(
    'security_group_instance_associations',
    meta,
    Column('id', String(length=36), primary_key=True, nullable=False),
    Column('security_group_id', String(length=36),
           ForeignKey('security_groups.id', ondelete="CASCADE",
                      onupdate="CASCADE")),
    Column('instance_id', String(length=36),
           ForeignKey('instances.id', ondelete="CASCADE",
                      onupdate="CASCADE")),
    Column('created', DateTime()),
    Column('updated', DateTime()),
    Column('deleted', Boolean(), default=0),
    Column('deleted_at', DateTime()),
)

security_group_rules = Table(
    'security_group_rules',
    meta,
    Column('id', String(length=36), primary_key=True, nullable=False),
    Column('group_id', String(length=36),
           ForeignKey('security_groups.id', ondelete="CASCADE",
                      onupdate="CASCADE")),
    Column('parent_group_id', String(length=36),
           ForeignKey('security_groups.id', ondelete="CASCADE",
                      onupdate="CASCADE")),
    Column('protocol', String(length=255)),
    Column('from_port', Integer()),
    Column('to_port', Integer()),
    Column('cidr', String(length=255)),
    Column('created', DateTime()),
    Column('updated', DateTime()),
    Column('deleted', Boolean(), default=0),
    Column('deleted_at', DateTime()),
)


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    Table(
        'instances',
        meta,
        autoload=True,
    )
    create_tables([security_groups, security_group_rules,
                   security_group_instance_associations])



# Copyright 2016 Tesora, Inc.
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_serialization import jsonutils

from trove.common import crypto_utils as cu
from trove.common.rpc import serializer
from trove.instance.models import get_instance_encryption_key

CONF = cfg.CONF


# BUG(1650518): Cleanup in the Pike release
class ConductorHostSerializer(serializer.TroveSerializer):
    def __init__(self, base, *_):
        super(ConductorHostSerializer, self).__init__(base)

    def _serialize_entity(self, ctxt, entity):
        try:
            if ctxt.instance_id is None:
                return entity
        except (ValueError, TypeError):
            return entity

        instance_key = get_instance_encryption_key(ctxt.instance_id)

        estr = jsonutils.dumps(entity)
        return cu.encode_data(cu.encrypt_data(estr, instance_key))

    def _deserialize_entity(self, ctxt, entity):
        try:
            entity = jsonutils.loads(entity)
            instance_id = entity['csz-instance-id']
        except (ValueError, TypeError):
            return entity

        instance_key = get_instance_encryption_key(instance_id)

        estr = cu.decrypt_data(cu.decode_data(entity['entity']),
                               instance_key)
        entity = jsonutils.loads(estr)

        return entity

    def _serialize_context(self, ctxt):
        try:
            if ctxt.instance_id is None:
                return ctxt
        except (ValueError, TypeError):
            return ctxt

        instance_key = get_instance_encryption_key(ctxt.instance_id)

        cstr = jsonutils.dumps(ctxt)
        return {'context': cu.encode_data(cu.encrypt_data(cstr,
                                                          instance_key))}

    def _deserialize_context(self, ctxt):
        try:
            instance_id = ctxt.get('csz-instance-id', None)

            if instance_id is not None:
                instance_key = get_instance_encryption_key(instance_id)

                cstr = cu.decrypt_data(cu.decode_data(ctxt['context']),
                                       instance_key)
                ctxt = jsonutils.loads(cstr)
        except (ValueError, TypeError):
            return ctxt

        ctxt['instance_id'] = instance_id
        return ctxt



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from unittest import mock

from trove.common import cfg
from trove.common import wsgi
from trove.configuration import models as config_models
from trove.configuration import service
from trove.datastore import models as ds_models
from trove.instance import models as ins_models
from trove.tests.unittests import trove_testtools
from trove.tests.unittests.util import util

CONF = cfg.CONF


class TestConfigurationsController(trove_testtools.TestCase):
    @classmethod
    def setUpClass(cls):
        util.init_db()

        cls.ds_name = cls.random_name(
            'datastore', prefix='TestConfigurationsController')
        ds_models.update_datastore(name=cls.ds_name, default_version=None)
        cls.ds = ds_models.Datastore.load(cls.ds_name)

        ds_version_name = cls.random_name(
            'version', prefix='TestConfigurationsController')
        ds_models.update_datastore_version(
            cls.ds_name, ds_version_name, 'mysql', '',
            ['trove'], '', 1, version='5.7.29')
        cls.ds_version = ds_models.DatastoreVersion.load(
            cls.ds, ds_version_name, version='5.7.29')

        cls.tenant_id = cls.random_uuid()
        cls.config = config_models.Configuration.create(
            cls.random_name('configuration'),
            '', cls.tenant_id, None,
            cls.ds_version.id)
        cls.config_id = cls.config.id

        cls.controller = service.ConfigurationsController()

        super(TestConfigurationsController, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        util.cleanup_db()

        super(TestConfigurationsController, cls).tearDownClass()

    def test_show(self):
        req_mock = mock.MagicMock(
            environ={
                wsgi.CONTEXT_KEY: mock.MagicMock(
                    project_id=self.tenant_id,
                    is_admin=False
                )
            }
        )

        result = self.controller.show(req_mock, self.tenant_id,
                                      self.config_id)
        data = result.data(None).get('configuration')

        expected = {
            "id": self.config_id,
            "name": self.config.name,
            "description": '',
            "instance_count": 0,
            "datastore_name": self.ds_name,
            "datastore_version_id": self.ds_version.id,
            "datastore_version_name": self.ds_version.name,
            "datastore_version_number": self.ds_version.version
        }

        self.assertDictContains(data, expected)

    def test_instances_admin(self):
        """Admin user can get instances of specified config.

        Even the instance belongs to other user.
        """
        admin_project_id = self.random_uuid()
        req_mock = mock.MagicMock(
            environ={
                wsgi.CONTEXT_KEY: mock.MagicMock(
                    project_id=admin_project_id,
                    is_admin=True,
                    limit=None,
                    marker=None
                )
            }
        )

        # Create a config
        config = config_models.Configuration.create(
            self.random_name('configuration'),
            '', self.tenant_id, None,
            self.ds_version.id)

        # Create an instance
        ins_name = self.random_name('instance')
        instance = ins_models.DBInstance.create(
            name=ins_name,
            flavor_id=self.random_uuid(),
            tenant_id=self.tenant_id,
            volume_size=1,
            datastore_version_id=self.ds_version.id,
            task_status=ins_models.InstanceTasks.NONE,
            compute_instance_id=self.random_uuid(),
            server_status='ACTIVE',
            configuration_id=config.id
        )

        result = self.controller.instances(req_mock, admin_project_id,
                                           config.id)

        self.assertEqual(200, result.status)
        instances = result.data(None)['instances']
        self.assertEqual(1, len(instances))
        self.assertEqual({'id': instance.id, 'name': ins_name}, instances[0])



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.cluster_group"


class ClusterRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'cluster_runners'
    _runner_cls = 'ClusterRunner'


@test(groups=[GROUP, groups.CLUSTER_CFGGRP_CREATE],
      runs_after_groups=[groups.MODULE_DELETE,
                         groups.CFGGRP_INST_DELETE,
                         groups.INST_ACTIONS_RESIZE_WAIT,
                         groups.DB_ACTION_INST_DELETE,
                         groups.USER_ACTION_DELETE,
                         groups.USER_ACTION_INST_DELETE,
                         groups.ROOT_ACTION_INST_DELETE,
                         groups.REPL_INST_DELETE_WAIT,
                         groups.INST_DELETE])
class ClusterConfigurationCreateGroup(TestGroup):

    def __init__(self):
        super(ClusterConfigurationCreateGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def create_initial_configuration(self):
        """Create a configuration group for a new cluster."""
        self.test_runner.run_initial_configuration_create()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS, groups.CLUSTER_CREATE],
      runs_after_groups=[groups.CLUSTER_CFGGRP_CREATE])
class ClusterCreateGroup(TestGroup):

    def __init__(self):
        super(ClusterCreateGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_create(self):
        """Create a cluster."""
        self.test_runner.run_cluster_create()


@test(groups=[GROUP, groups.CLUSTER_CREATE_WAIT],
      depends_on_groups=[groups.CLUSTER_CREATE],
      runs_after_groups=[groups.MODULE_INST_DELETE_WAIT,
                         groups.CFGGRP_INST_DELETE_WAIT,
                         groups.DB_ACTION_INST_DELETE_WAIT,
                         groups.USER_ACTION_INST_DELETE_WAIT,
                         groups.ROOT_ACTION_INST_DELETE_WAIT,
                         groups.INST_DELETE_WAIT])
class ClusterCreateWaitGroup(TestGroup):

    def __init__(self):
        super(ClusterCreateWaitGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_create_wait(self):
        """Wait for cluster create to complete."""
        self.test_runner.run_cluster_create_wait()

    @test(depends_on=[cluster_create_wait])
    def verify_initial_configuration(self):
        """Verify initial configuration values on the cluster."""
        self.test_runner.run_verify_initial_configuration()

    @test(depends_on=[cluster_create_wait])
    def add_initial_cluster_data(self):
        """Add data to cluster."""
        self.test_runner.run_add_initial_cluster_data()

    @test(depends_on=[add_initial_cluster_data])
    def verify_initial_cluster_data(self):
        """Verify the initial data exists on cluster."""
        self.test_runner.run_verify_initial_cluster_data()

    @test(depends_on=[cluster_create_wait])
    def cluster_list(self):
        """List the clusters."""
        self.test_runner.run_cluster_list()

    @test(depends_on=[cluster_create_wait])
    def cluster_show(self):
        """Show a cluster."""
        self.test_runner.run_cluster_show()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_RESTART],
      depends_on_groups=[groups.CLUSTER_CREATE_WAIT])
class ClusterRestartGroup(TestGroup):

    def __init__(self):
        super(ClusterRestartGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_restart(self):
        """Restart the cluster."""
        self.test_runner.run_cluster_restart()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_RESTART_WAIT],
      depends_on_groups=[groups.CLUSTER_ACTIONS_RESTART])
class ClusterRestartWaitGroup(TestGroup):

    def __init__(self):
        super(ClusterRestartWaitGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_restart_wait(self):
        """Wait for cluster restart to complete."""
        self.test_runner.run_cluster_restart_wait()

    @test(depends_on=[cluster_restart_wait])
    def verify_initial_cluster_data(self):
        """Verify the initial data still exists after cluster restart."""
        self.test_runner.run_verify_initial_cluster_data()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_ROOT_ENABLE],
      depends_on_groups=[groups.CLUSTER_CREATE_WAIT],
      runs_after_groups=[groups.CLUSTER_ACTIONS_RESTART_WAIT])
class ClusterRootEnableGroup(TestGroup):

    def __init__(self):
        super(ClusterRootEnableGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_root_enable(self):
        """Root Enable."""
        self.test_runner.run_cluster_root_enable()

    @test(depends_on=[cluster_root_enable])
    def verify_cluster_root_enable(self):
        """Verify Root Enable."""
        self.test_runner.run_verify_cluster_root_enable()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_GROW_SHRINK,
              groups.CLUSTER_ACTIONS_GROW],
      depends_on_groups=[groups.CLUSTER_CREATE_WAIT],
      runs_after_groups=[groups.CLUSTER_ACTIONS_ROOT_ENABLE])
class ClusterGrowGroup(TestGroup):

    def __init__(self):
        super(ClusterGrowGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_grow(self):
        """Grow cluster."""
        self.test_runner.run_cluster_grow()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_GROW_SHRINK,
              groups.CLUSTER_ACTIONS_GROW_WAIT],
      depends_on_groups=[groups.CLUSTER_ACTIONS_GROW])
class ClusterGrowWaitGroup(TestGroup):
    def __init__(self):
        super(ClusterGrowWaitGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_grow_wait(self):
        """Wait for cluster grow to complete."""
        self.test_runner.run_cluster_grow_wait()

    @test(depends_on=[cluster_grow_wait])
    def verify_initial_configuration(self):
        """Verify initial configuration values on the cluster."""
        self.test_runner.run_verify_initial_configuration()

    @test(depends_on=[cluster_grow_wait])
    def verify_initial_cluster_data_after_grow(self):
        """Verify the initial data still exists after cluster grow."""
        self.test_runner.run_verify_initial_cluster_data()

    @test(depends_on=[cluster_grow_wait],
          runs_after=[verify_initial_cluster_data_after_grow])
    def add_grow_cluster_data(self):
        """Add more data to cluster after grow."""
        self.test_runner.run_add_grow_cluster_data()

    @test(depends_on=[add_grow_cluster_data])
    def verify_grow_cluster_data(self):
        """Verify the data added after cluster grow."""
        self.test_runner.run_verify_grow_cluster_data()

    @test(depends_on=[add_grow_cluster_data],
          runs_after=[verify_grow_cluster_data])
    def remove_grow_cluster_data(self):
        """Remove the data added after cluster grow."""
        self.test_runner.run_remove_grow_cluster_data()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_ROOT_ACTIONS,
              groups.CLUSTER_ACTIONS_ROOT_GROW],
      depends_on_groups=[groups.CLUSTER_ACTIONS_GROW_WAIT])
class ClusterRootEnableGrowGroup(TestGroup):

    def __init__(self):
        super(ClusterRootEnableGrowGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def verify_cluster_root_enable_after_grow(self):
        """Verify Root Enabled after grow."""
        self.test_runner.run_verify_cluster_root_enable()


@test(groups=[GROUP, groups.CLUSTER_UPGRADE],
      depends_on_groups=[groups.CLUSTER_CREATE_WAIT],
      runs_after_groups=[groups.CLUSTER_ACTIONS_GROW_WAIT,
                         groups.CLUSTER_ACTIONS_ROOT_GROW])
class ClusterUpgradeGroup(TestGroup):

    def __init__(self):
        super(ClusterUpgradeGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_upgrade(self):
        """Upgrade cluster."""
        self.test_runner.run_cluster_upgrade()


@test(groups=[GROUP, groups.CLUSTER_UPGRADE_WAIT],
      depends_on_groups=[groups.CLUSTER_UPGRADE])
class ClusterUpgradeWaitGroup(TestGroup):
    def __init__(self):
        super(ClusterUpgradeWaitGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_upgrade_wait(self):
        """Wait for cluster upgrade to complete."""
        self.test_runner.run_cluster_upgrade_wait()

    @test(depends_on=[cluster_upgrade_wait])
    def verify_initial_configuration(self):
        """Verify initial configuration values on the cluster."""
        self.test_runner.run_verify_initial_configuration()

    @test(depends_on=[cluster_upgrade_wait])
    def verify_initial_cluster_data_after_upgrade(self):
        """Verify the initial data still exists after cluster upgrade."""
        self.test_runner.run_verify_initial_cluster_data()

    @test(depends_on=[cluster_upgrade_wait],
          runs_after=[verify_initial_cluster_data_after_upgrade])
    def add_upgrade_cluster_data_after_upgrade(self):
        """Add more data to cluster after upgrade."""
        self.test_runner.run_add_upgrade_cluster_data()

    @test(depends_on=[add_upgrade_cluster_data_after_upgrade])
    def verify_upgrade_cluster_data_after_upgrade(self):
        """Verify the data added after cluster upgrade."""
        self.test_runner.run_verify_upgrade_cluster_data()

    @test(depends_on=[add_upgrade_cluster_data_after_upgrade],
          runs_after=[verify_upgrade_cluster_data_after_upgrade])
    def remove_upgrade_cluster_data_after_upgrade(self):
        """Remove the data added after cluster upgrade."""
        self.test_runner.run_remove_upgrade_cluster_data()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_GROW_SHRINK,
              groups.CLUSTER_ACTIONS_SHRINK],
      depends_on_groups=[groups.CLUSTER_ACTIONS_GROW_WAIT],
      runs_after_groups=[groups.CLUSTER_UPGRADE_WAIT])
class ClusterShrinkGroup(TestGroup):

    def __init__(self):
        super(ClusterShrinkGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_shrink(self):
        """Shrink cluster."""
        self.test_runner.run_cluster_shrink()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_SHRINK_WAIT],
      depends_on_groups=[groups.CLUSTER_ACTIONS_SHRINK])
class ClusterShrinkWaitGroup(TestGroup):
    def __init__(self):
        super(ClusterShrinkWaitGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_shrink_wait(self):
        """Wait for the cluster shrink to complete."""
        self.test_runner.run_cluster_shrink_wait()

    @test(depends_on=[cluster_shrink_wait])
    def verify_initial_configuration(self):
        """Verify initial configuration values on the cluster."""
        self.test_runner.run_verify_initial_configuration()

    @test(depends_on=[cluster_shrink_wait])
    def verify_initial_cluster_data_after_shrink(self):
        """Verify the initial data still exists after cluster shrink."""
        self.test_runner.run_verify_initial_cluster_data()

    @test(runs_after=[verify_initial_cluster_data_after_shrink])
    def add_shrink_cluster_data(self):
        """Add more data to cluster after shrink."""
        self.test_runner.run_add_shrink_cluster_data()

    @test(depends_on=[add_shrink_cluster_data])
    def verify_shrink_cluster_data(self):
        """Verify the data added after cluster shrink."""
        self.test_runner.run_verify_shrink_cluster_data()

    @test(depends_on=[add_shrink_cluster_data],
          runs_after=[verify_shrink_cluster_data])
    def remove_shrink_cluster_data(self):
        """Remove the data added after cluster shrink."""
        self.test_runner.run_remove_shrink_cluster_data()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_ROOT_ACTIONS,
              groups.CLUSTER_ACTIONS_ROOT_SHRINK],
      depends_on_groups=[groups.CLUSTER_ACTIONS_SHRINK_WAIT])
class ClusterRootEnableShrinkGroup(TestGroup):

    def __init__(self):
        super(ClusterRootEnableShrinkGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def verify_cluster_root_enable_after_shrink(self):
        """Verify Root Enable after shrink."""
        self.test_runner.run_verify_cluster_root_enable()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_ACTIONS_CFGGRP_ACTIONS],
      depends_on_groups=[groups.CLUSTER_CREATE_WAIT],
      runs_after_groups=[groups.CLUSTER_ACTIONS_ROOT_SHRINK])
class ClusterConfigurationActionsGroup(TestGroup):

    def __init__(self):
        super(ClusterConfigurationActionsGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def detach_initial_configuration(self):
        """Detach initial configuration group."""
        self.test_runner.run_detach_initial_configuration()

    @test(depends_on=[detach_initial_configuration])
    def restart_cluster_after_detach(self):
        """Restarting cluster after configuration change."""
        self.test_runner.restart_after_configuration_change()

    @test
    def create_dynamic_configuration(self):
        """Create a configuration group with only dynamic entries."""
        self.test_runner.run_create_dynamic_configuration()

    @test
    def create_non_dynamic_configuration(self):
        """Create a configuration group with only non-dynamic entries."""
        self.test_runner.run_create_non_dynamic_configuration()

    @test(depends_on=[create_dynamic_configuration,
                      restart_cluster_after_detach])
    def attach_dynamic_configuration(self):
        """Test attach dynamic group."""
        self.test_runner.run_attach_dynamic_configuration()

    @test(depends_on=[attach_dynamic_configuration])
    def verify_dynamic_configuration(self):
        """Verify dynamic values on the cluster."""
        self.test_runner.run_verify_dynamic_configuration()

    @test(depends_on=[attach_dynamic_configuration],
          runs_after=[verify_dynamic_configuration])
    def detach_dynamic_configuration(self):
        """Test detach dynamic group."""
        self.test_runner.run_detach_dynamic_configuration()

    @test(depends_on=[create_non_dynamic_configuration,
                      detach_initial_configuration],
          runs_after=[detach_dynamic_configuration])
    def attach_non_dynamic_configuration(self):
        """Test attach non-dynamic group."""
        self.test_runner.run_attach_non_dynamic_configuration()

    @test(depends_on=[attach_non_dynamic_configuration])
    def restart_cluster_after_attach(self):
        """Restarting cluster after configuration change."""
        self.test_runner.restart_after_configuration_change()

    @test(depends_on=[restart_cluster_after_attach])
    def verify_non_dynamic_configuration(self):
        """Verify non-dynamic values on the cluster."""
        self.test_runner.run_verify_non_dynamic_configuration()

    @test(depends_on=[attach_non_dynamic_configuration],
          runs_after=[verify_non_dynamic_configuration])
    def detach_non_dynamic_configuration(self):
        """Test detach non-dynamic group."""
        self.test_runner.run_detach_non_dynamic_configuration()

    @test(runs_after=[detach_dynamic_configuration,
                      detach_non_dynamic_configuration])
    def verify_initial_cluster_data(self):
        """Verify the initial data still exists."""
        self.test_runner.run_verify_initial_cluster_data()

    @test(depends_on=[detach_dynamic_configuration])
    def delete_dynamic_configuration(self):
        """Test delete dynamic configuration group."""
        self.test_runner.run_delete_dynamic_configuration()

    @test(depends_on=[detach_non_dynamic_configuration])
    def delete_non_dynamic_configuration(self):
        """Test delete non-dynamic configuration group."""
        self.test_runner.run_delete_non_dynamic_configuration()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_DELETE],
      depends_on_groups=[groups.CLUSTER_CREATE_WAIT],
      runs_after_groups=[groups.CLUSTER_ACTIONS_ROOT_ENABLE,
                         groups.CLUSTER_ACTIONS_ROOT_GROW,
                         groups.CLUSTER_ACTIONS_ROOT_SHRINK,
                         groups.CLUSTER_ACTIONS_GROW_WAIT,
                         groups.CLUSTER_ACTIONS_SHRINK_WAIT,
                         groups.CLUSTER_UPGRADE_WAIT,
                         groups.CLUSTER_ACTIONS_RESTART_WAIT,
                         groups.CLUSTER_CFGGRP_CREATE,
                         groups.CLUSTER_ACTIONS_CFGGRP_ACTIONS])
class ClusterDeleteGroup(TestGroup):

    def __init__(self):
        super(ClusterDeleteGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def remove_initial_cluster_data(self):
        """Remove the initial data from cluster."""
        self.test_runner.run_remove_initial_cluster_data()

    @test(runs_after=[remove_initial_cluster_data])
    def cluster_delete(self):
        """Delete an existing cluster."""
        self.test_runner.run_cluster_delete()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_DELETE_WAIT],
      depends_on_groups=[groups.CLUSTER_DELETE])
class ClusterDeleteWaitGroup(TestGroup):

    def __init__(self):
        super(ClusterDeleteWaitGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def cluster_delete_wait(self):
        """Wait for the existing cluster to be gone."""
        self.test_runner.run_cluster_delete_wait()


@test(groups=[GROUP, groups.CLUSTER_ACTIONS,
              groups.CLUSTER_CFGGRP_DELETE],
      depends_on_groups=[groups.CLUSTER_CFGGRP_CREATE],
      runs_after_groups=[groups.CLUSTER_DELETE_WAIT])
class ClusterConfigurationDeleteGroup(TestGroup):

    def __init__(self):
        super(ClusterConfigurationDeleteGroup, self).__init__(
            ClusterRunnerFactory.instance())

    @test
    def delete_initial_configuration(self):
        """Delete initial configuration group."""
        self.test_runner.run_delete_initial_configuration()



#    Copyright 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nose.plugins.skip import SkipTest
import proboscis
from proboscis.asserts import assert_equal
from proboscis.asserts import assert_false
from proboscis.asserts import assert_not_equal
from proboscis.asserts import assert_raises
from proboscis.asserts import assert_true
from proboscis import test
from troveclient.compat import exceptions

from trove import tests
from trove.tests.api import instances
from trove.tests.util import test_config


@test(groups=[tests.DBAAS_API_USERS_ROOT],
      depends_on_groups=[tests.DBAAS_API_INSTANCES])
class TestRoot(object):
    root_enabled_timestamp = 'Never'

    @proboscis.before_class
    def setUp(self):
        # Reuse the instance created previously.
        self.id = instances.instance_info.id
        self.dbaas = instances.instance_info.dbaas
        self.dbaas_admin = instances.instance_info.dbaas_admin

    def _verify_root_timestamp(self, id):
        reh = self.dbaas_admin.management.root_enabled_history(id)
        timestamp = reh.enabled
        assert_equal(self.root_enabled_timestamp, timestamp)
        assert_equal(id, reh.id)

    def _root(self):
        self.dbaas.root.create(self.id)
        assert_equal(200, self.dbaas.last_http_code)
        reh = self.dbaas_admin.management.root_enabled_history
        self.root_enabled_timestamp = reh(self.id).enabled

    @test
    def test_root_initially_disabled(self):
        """Test that root is disabled."""
        enabled = self.dbaas.root.is_root_enabled(self.id)
        assert_equal(200, self.dbaas.last_http_code)

        is_enabled = enabled
        if hasattr(enabled, 'rootEnabled'):
            is_enabled = enabled.rootEnabled
        assert_false(is_enabled, "Root SHOULD NOT be enabled.")

    @test
    def test_create_user_os_admin_failure(self):
        users = [{"name": "os_admin", "password": "12345"}]
        assert_raises(exceptions.BadRequest, self.dbaas.users.create,
                      self.id, users)

    @test
    def test_delete_user_os_admin_failure(self):
        assert_raises(exceptions.BadRequest, self.dbaas.users.delete,
                      self.id, "os_admin")

    @test(depends_on=[test_root_initially_disabled],
          enabled=not test_config.values['root_removed_from_instance_api'])
    def test_root_initially_disabled_details(self):
        """Use instance details to test that root is disabled."""
        instance = self.dbaas.instances.get(self.id)
        assert_true(hasattr(instance, 'rootEnabled'),
                    "Instance has no rootEnabled property.")
        assert_false(instance.rootEnabled, "Root SHOULD NOT be enabled.")
        assert_equal(self.root_enabled_timestamp, 'Never')

    @test(depends_on=[test_root_initially_disabled_details])
    def test_root_disabled_in_mgmt_api(self):
        """Verifies in the management api that the timestamp exists."""
        self._verify_root_timestamp(self.id)

    @test(depends_on=[test_root_initially_disabled_details])
    def test_root_disable_when_root_not_enabled(self):
        reh = self.dbaas_admin.management.root_enabled_history
        self.root_enabled_timestamp = reh(self.id).enabled
        assert_raises(exceptions.NotFound, self.dbaas.root.delete,
                      self.id)
        self._verify_root_timestamp(self.id)

    @test(depends_on=[test_root_disable_when_root_not_enabled])
    def test_enable_root(self):
        self._root()

    @test(depends_on=[test_enable_root])
    def test_enabled_timestamp(self):
        assert_not_equal(self.root_enabled_timestamp, 'Never')

    @test(depends_on=[test_enable_root])
    def test_root_not_in_users_list(self):
        """
        Tests that despite having enabled root, user root doesn't appear
        in the users list for the instance.
        """
        users = self.dbaas.users.list(self.id)
        usernames = [user.name for user in users]
        assert_true('root' not in usernames)

    @test(depends_on=[test_enable_root])
    def test_root_now_enabled(self):
        """Test that root is now enabled."""
        enabled = self.dbaas.root.is_root_enabled(self.id)
        assert_equal(200, self.dbaas.last_http_code)
        assert_true(enabled, "Root SHOULD be enabled.")

    @test(depends_on=[test_root_now_enabled],
          enabled=not test_config.values['root_removed_from_instance_api'])
    def test_root_now_enabled_details(self):
        """Use instance details to test that root is now enabled."""
        instance = self.dbaas.instances.get(self.id)
        assert_true(hasattr(instance, 'rootEnabled'),
                    "Instance has no rootEnabled property.")
        assert_true(instance.rootEnabled, "Root SHOULD be enabled.")
        assert_not_equal(self.root_enabled_timestamp, 'Never')
        self._verify_root_timestamp(self.id)

    @test(depends_on=[test_root_now_enabled_details])
    def test_reset_root(self):
        if test_config.values['root_timestamp_disabled']:
            raise SkipTest("Enabled timestamp not enabled yet")
        old_ts = self.root_enabled_timestamp
        self._root()
        assert_not_equal(self.root_enabled_timestamp, 'Never')
        assert_equal(self.root_enabled_timestamp, old_ts)

    @test(depends_on=[test_reset_root])
    def test_root_still_enabled(self):
        """Test that after root was reset it's still enabled."""
        enabled = self.dbaas.root.is_root_enabled(self.id)
        assert_equal(200, self.dbaas.last_http_code)
        assert_true(enabled, "Root SHOULD still be enabled.")

    @test(depends_on=[test_root_still_enabled],
          enabled=not test_config.values['root_removed_from_instance_api'])
    def test_root_still_enabled_details(self):
        """Use instance details to test that after root was reset,
            it's still enabled.
        """
        instance = self.dbaas.instances.get(self.id)
        assert_true(hasattr(instance, 'rootEnabled'),
                    "Instance has no rootEnabled property.")
        assert_true(instance.rootEnabled, "Root SHOULD still be enabled.")
        assert_not_equal(self.root_enabled_timestamp, 'Never')
        self._verify_root_timestamp(self.id)

    @test(depends_on=[test_enable_root])
    def test_root_cannot_be_deleted(self):
        """Even if root was enabled, the user root cannot be deleted."""
        assert_raises(exceptions.BadRequest, self.dbaas.users.delete,
                      self.id, "root")

    @test(depends_on=[test_root_still_enabled_details])
    def test_root_disable(self):
        reh = self.dbaas_admin.management.root_enabled_history
        self.root_enabled_timestamp = reh(self.id).enabled
        self.dbaas.root.delete(self.id)
        assert_equal(204, self.dbaas.last_http_code)
        self._verify_root_timestamp(self.id)



# Copyright 2015 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_context import context
from oslo_log import log as logging
import oslo_messaging as messaging
from osprofiler import notifier
from osprofiler import web

from trove.common import cfg
from trove import rpc


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def setup_profiler(binary, host):
    if CONF.profiler.enabled:
        _notifier = notifier.create(
            "Messaging", messaging, context.get_admin_context().to_dict(),
            rpc.TRANSPORT, "trove", binary, host)
        notifier.set(_notifier)
        web.enable(CONF.profiler.hmac_keys)
        LOG.warning("The OpenStack Profiler is enabled. Using one"
                    " of the hmac_keys specified in the trove.conf file "
                    "(typically in /etc/trove), a trace can be made of "
                    "all requests. Only an admin user can retrieve "
                    "the trace information, however.\n"
                    "To disable the profiler, add the following to the "
                    "configuration file:\n"
                    "[profiler]\n"
                    "enabled=false")
    else:
        web.disable()



#    Copyright (c) 2014 Rackspace Hosting
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import ForeignKey
from sqlalchemy.schema import Column
from sqlalchemy.schema import MetaData
from sqlalchemy.schema import UniqueConstraint

from trove.db.sqlalchemy.migrate_repo.schema import Boolean
from trove.db.sqlalchemy.migrate_repo.schema import create_tables
from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table


meta = MetaData()

capabilities = Table(
    'capabilities',
    meta,
    Column('id', String(36), primary_key=True, nullable=False),
    Column('name', String(255), unique=True),
    Column('description', String(255), nullable=False),
    Column('enabled', Boolean())
)


capability_overrides = Table(
    'capability_overrides',
    meta,
    Column('id', String(36), primary_key=True, nullable=False),
    Column('datastore_version_id', String(36),
           ForeignKey('datastore_versions.id')),
    Column('capability_id', String(36), ForeignKey('capabilities.id')),
    Column('enabled', Boolean()),
    UniqueConstraint('datastore_version_id', 'capability_id',
                     name='idx_datastore_capabilities_enabled')
)


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    Table('datastores', meta, autoload=True)
    Table('datastore_versions', meta, autoload=True)
    create_tables([capabilities, capability_overrides])



# Copyright 2010-2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class AuthProtocol(object):

    def __init__(self, app, conf):
        self.conf = conf
        self.app = app

    def __call__(self, env, start_response):
        token = self._get_user_token_from_header(env)
        user_headers = self._get_info_from_token(token)
        self._add_headers(env, user_headers)
        return self.app(env, start_response)

    def _header_to_env_var(self, key):
        """Convert header to wsgi env variable.

        :param key: http header name (ex. 'X-Auth-Token')
        :return: wsgi env variable name (ex. 'HTTP_X_AUTH_TOKEN')

        """
        return 'HTTP_%s' % key.replace('-', '_').upper()

    def _add_headers(self, env, headers):
        """Add http headers to environment."""
        for (k, v) in headers.items():
            env_key = self._header_to_env_var(k)
            env[env_key] = v

    def get_admin_token(self):
        return "ABCDEF0123456789"

    def _get_info_from_token(self, token):
        if token.startswith("admin"):
            role = "admin,%s" % token
        else:
            role = token
        return {
            'X_IDENTITY_STATUS': 'Confirmed',
            'X_TENANT_ID': token,
            'X_TENANT_NAME': token,
            'X_USER_ID': token,
            'X_USER_NAME': token,
            'X_ROLE': role,
        }

    def _get_header(self, env, key, default=None):
        # Copied from keystone.
        env_key = self._header_to_env_var(key)
        return env.get(env_key, default)

    def _get_user_token_from_header(self, env):
        token = self._get_header(env, 'X-Auth-Token',
                                 self._get_header(env, 'X-Storage-Token'))
        if token:
            return token
        else:
            raise RuntimeError('Unable to find token in headers')


def filter_factory(global_conf, **local_conf):
    """Fakes a keystone filter."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def auth_filter(app):
        return AuthProtocol(app, conf)
    return auth_filter



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest.mock import Mock

from trove.common import pagination
from trove.tests.unittests import trove_testtools


class TestPaginatedDataView(trove_testtools.TestCase):

    def test_creation_with_string_marker(self):
        view = pagination.PaginatedDataView("TestType", [],
                                            "http://current_page",
                                            next_page_marker="marker")
        self.assertEqual("marker", view.next_page_marker)

    def test_creation_with_none_marker(self):
        view = pagination.PaginatedDataView("TestType", [],
                                            "http://current_page",
                                            next_page_marker=None)
        self.assertIsNone(view.next_page_marker)

    def test_creation_with_none_string_marker(self):
        view = pagination.PaginatedDataView("TestType", [],
                                            "http://current_page",
                                            next_page_marker=52)
        self.assertEqual("52", view.next_page_marker)

    def _do_paginate_list(self, limit=None, marker='', include_marker=False):
        li = ['a', 'b', 'c', 'd', 'e']
        return pagination.paginate_list(li, limit=limit, marker=marker,
                                        include_marker=include_marker)

    def test_paginate_list(self):
        # start list
        li_1, marker_1 = self._do_paginate_list(limit=2)
        self.assertEqual(['a', 'b'], li_1)
        self.assertEqual('b', marker_1)

        # continue list, do not include marker in result
        li_2, marker_2 = self._do_paginate_list(limit=2, marker=marker_1)
        self.assertEqual(['c', 'd'], li_2)
        self.assertEqual('d', marker_2)
        li_3, marker_3 = self._do_paginate_list(limit=2, marker=marker_2)
        self.assertEqual(['e'], li_3)
        self.assertIsNone(marker_3)

        # alternate continue list, include marker in result
        li_4, marker_4 = self._do_paginate_list(limit=2, marker=marker_1,
                                                include_marker=True)
        self.assertEqual(['b', 'c'], li_4)
        self.assertEqual('c', marker_4)
        li_5, marker_5 = self._do_paginate_list(limit=2, marker=marker_4,
                                                include_marker=True)
        self.assertEqual(['c', 'd'], li_5)
        self.assertEqual('d', marker_5)
        li_6, marker_6 = self._do_paginate_list(limit=2, marker=marker_5,
                                                include_marker=True)
        self.assertEqual(['d', 'e'], li_6)
        self.assertIsNone(marker_6)

        # bad marker
        li_4, marker_4 = self._do_paginate_list(marker='f')
        self.assertEqual([], li_4)
        self.assertIsNone(marker_4)

        li_5, marker_5 = self._do_paginate_list(limit=1, marker='f')
        self.assertEqual([], li_5)
        self.assertIsNone(marker_5)

    def test_dict_paginate(self):
        li = [{'_collate': 'en_US.UTF-8',
               '_character_set': 'UTF8',
               '_name': 'db1'},
              {'_collate': 'en_US.UTF-8',
               '_character_set': 'UTF8',
               '_name': 'db3'},
              {'_collate': 'en_US.UTF-8',
               '_character_set': 'UTF8',
               '_name': 'db2'},
              {'_collate': 'en_US.UTF-8',
               '_character_set': 'UTF8',
               '_name': 'db5'},
              {'_collate': 'en_US.UTF-8',
               '_character_set': 'UTF8',
               '_name': 'db4'}
              ]

        l, m = pagination.paginate_dict_list(li, '_name', limit=1,
                                             marker='db1',
                                             include_marker=True)
        self.assertEqual(l[0], li[0])
        self.assertEqual(m, 'db1')

    def test_object_paginate(self):

        def build_mock_object(name):
            o = Mock()
            o.name = name
            o.attr = 'attr'
            return o

        li = [build_mock_object('db1'), build_mock_object('db2'),
              build_mock_object('db3')]

        l, m = pagination.paginate_object_list(li, 'name', limit=1,
                                               marker='db1',
                                               include_marker=True)
        self.assertEqual(l[0], li[0])
        self.assertEqual(m, 'db1')



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


class FakeNeutronClient(object):
    def __init__(self, context):
        self.context = context

    def show_network(self, *arg, **kwargs):
        return {'network': {'name': 'fake-mgmt-net-name'}}

    def list_networks(self, *arg, **kwargs):
        if 'router:external' in kwargs:
            return {'networks': [{'id': 'fake-public-net-id'}]}

        return {'networks': []}

    def create_port(self, body):
        if 'Management' in body['port'].get('description', ''):
            return {'port': {'id': 'fake-mgmt-port-id'}}

        return {'port': {'id': 'fake-user-port-id'}}

    def delete_port(self, *arg, **kwargs):
        pass

    def list_ports(self, *arg, **kwargs):
        return {'ports': []}

    def create_floatingip(self, *arg, **kwargs):
        pass

    def list_floatingips(self, *arg, **kwargs):
        return {'floatingips': []}

    def update_floatingip(self, *arg, **kwargs):
        pass

    def delete_floatingip(self, *arg, **kwargs):
        pass

    def create_security_group(self, *arg, **kwargs):
        return {'security_group': {'id': 'fake-sg-id'}}

    def create_security_group_rule(self, *arg, **kwargs):
        pass

    def list_security_groups(self, *arg, **kwargs):
        return {'security_groups': []}

    def delete_security_group(self, *arg, **kwargs):
        pass


def fake_create_neutron_client(context, region_name=None):
    return FakeNeutronClient(context)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from trove.common.policies.base import (
    PATH_INSTANCES, PATH_INSTANCES_DETAIL, PATH_INSTANCE, PATH_INSTANCE_ACTION)


rules = [
    policy.DocumentedRuleDefault(
        name='instance:create',
        check_str='rule:admin_or_owner',
        description='Create a database instance.',
        operations=[
            {
                'path': PATH_INSTANCES,
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:delete',
        check_str='rule:admin_or_owner',
        description='Delete a database instance.',
        operations=[
            {
                'path': PATH_INSTANCE,
                'method': 'DELETE'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:force_delete',
        check_str='rule:admin_or_owner',
        description='Forcibly delete a database instance.',
        operations=[
            {
                'path': PATH_INSTANCE,
                'method': 'DELETE'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:index',
        check_str='rule:admin_or_owner',
        description='List database instances.',
        operations=[
            {
                'path': PATH_INSTANCES,
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:detail',
        check_str='rule:admin_or_owner',
        description='List database instances with details.',
        operations=[
            {
                'path': PATH_INSTANCES_DETAIL,
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:show',
        check_str='rule:admin_or_owner',
        description='Get details of a specific database instance.',
        operations=[
            {
                'path': PATH_INSTANCE,
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:update',
        check_str='rule:admin_or_owner',
        description='Update a database instance to '
                    'attach/detach configuration',
        operations=[
            {
                'path': PATH_INSTANCE,
                'method': 'PUT'
            },
            # we also check this when creating instances with
            # a configuration group specified.
            {
                'path': PATH_INSTANCES,
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:edit',
        check_str='rule:admin_or_owner',
        description='Updates the instance to set or '
                    'unset one or more attributes.',
        operations=[
            {
                'path': PATH_INSTANCE,
                'method': 'PATCH'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:restart',
        check_str='rule:admin_or_owner',
        description='Restart a database instance.',
        operations=[
            {
                'path': PATH_INSTANCE_ACTION + ' (restart)',
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:resize_volume',
        check_str='rule:admin_or_owner',
        description='Resize a database instance volume.',
        operations=[
            {
                'path': PATH_INSTANCE_ACTION + ' (resize)',
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:resize_flavor',
        check_str='rule:admin_or_owner',
        description='Resize a database instance flavor.',
        operations=[
            {
                'path': PATH_INSTANCE_ACTION + ' (resize)',
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:reset_status',
        check_str='rule:admin',
        description='Reset the status of a database instance to ERROR.',
        operations=[
            {
                'path': PATH_INSTANCE_ACTION + ' (reset_status)',
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:promote_to_replica_source',
        check_str='rule:admin_or_owner',
        description='Promote instance to replica source.',
        operations=[
            {
                'path': PATH_INSTANCE_ACTION + ' (promote_to_replica_source)',
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:eject_replica_source',
        check_str='rule:admin_or_owner',
        description='Eject the replica source from its replica set.',
        operations=[
            {
                'path': PATH_INSTANCE_ACTION + ' (eject_replica_source)',
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:configuration',
        check_str='rule:admin_or_owner',
        description='Get the default configuration template '
                    'applied to the instance.',
        operations=[
            {
                'path': PATH_INSTANCE + '/configuration',
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:guest_log_list',
        check_str='rule:admin_or_owner',
        description='Get all informations about all logs '
                    'of a database instance.',
        operations=[
            {
                'path': PATH_INSTANCE + '/log',
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:backups',
        check_str='rule:admin_or_owner',
        description='Get all backups of a database instance.',
        operations=[
            {
                'path': PATH_INSTANCE + '/backups',
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:module_list',
        check_str='rule:admin_or_owner',
        description='Get informations about modules on a database instance.',
        operations=[
            {
                'path': PATH_INSTANCE + '/modules',
                'method': 'GET'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:module_apply',
        check_str='rule:admin_or_owner',
        description='Apply modules to a database instance.',
        operations=[
            {
                'path': PATH_INSTANCE + '/modules',
                'method': 'POST'
            },
            # we also check this when creating instances with
            # modules specified.
            {
                'path': PATH_INSTANCES,
                'method': 'POST'
            }
        ]),
    policy.DocumentedRuleDefault(
        name='instance:module_remove',
        check_str='rule:admin_or_owner',
        description='Remove a module from a database instance.',
        operations=[
            {
                'path': PATH_INSTANCE + '/modules/{module_id}',
                'method': 'DELETE'
            }
        ])
]


def list_rules():
    return rules



# Copyright [2015] Hewlett-Packard Development Company, L.P.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from unittest.mock import Mock
from unittest.mock import patch
from novaclient import exceptions as nova_exceptions
from trove.cluster.models import Cluster
from trove.cluster.models import ClusterTasks
from trove.cluster.models import DBCluster
from trove.common import clients
from trove.common import exception
from trove.common.strategies.cluster.experimental.redis import api as redis_api
from trove.instance import models as inst_models
from trove.instance.models import DBInstance
from trove.instance.models import InstanceTasks
from trove.quota.quota import QUOTAS
from trove.taskmanager import api as task_api
from trove.tests.unittests import trove_testtools


class FakeOptGroup(object):
    def __init__(self, cluster_member_count=3,
                 volume_support=True, device_path='/dev/vdb'):
        self.cluster_member_count = cluster_member_count
        self.volume_support = volume_support
        self.device_path = device_path


class ClusterTest(trove_testtools.TestCase):
    def setUp(self):
        super(ClusterTest, self).setUp()

        self.cluster_id = str(uuid.uuid4())
        self.cluster_name = "Cluster" + self.cluster_id
        self.tenant_id = "23423432"
        self.dv_id = "1"
        self.db_info = DBCluster(ClusterTasks.NONE,
                                 id=self.cluster_id,
                                 name=self.cluster_name,
                                 tenant_id=self.tenant_id,
                                 datastore_version_id=self.dv_id,
                                 task_id=ClusterTasks.NONE._code)

        self.get_client_patch = patch.object(task_api.API, 'get_client')
        self.get_client_mock = self.get_client_patch.start()
        self.addCleanup(self.get_client_patch.stop)
        self.dbcreate_patch = patch.object(DBCluster, 'create',
                                           return_value=self.db_info)
        self.dbcreate_mock = self.dbcreate_patch.start()
        self.addCleanup(self.dbcreate_patch.stop)

        self.context = trove_testtools.TroveTestContext(self)
        self.datastore = Mock()
        self.dv = Mock()
        self.dv.manager = "redis"
        self.datastore_version = self.dv
        self.cluster = redis_api.RedisCluster(self.context, self.db_info,
                                              self.datastore,
                                              self.datastore_version)
        self.cluster._server_group_loaded = True
        self.instances_w_volumes = [{'volume_size': 1,
                                     'flavor_id': '1234'}] * 3
        self.instances_no_volumes = [{'flavor_id': '1234'}] * 3

    def tearDown(self):
        super(ClusterTest, self).tearDown()

    @patch.object(clients, 'create_nova_client')
    def test_create_invalid_flavor_specified(self,
                                             mock_client):
        (mock_client.return_value.flavors.get) = Mock(
            side_effect=nova_exceptions.NotFound(
                404, "Flavor id not found %s" % id))

        self.assertRaises(exception.FlavorNotFound,
                          Cluster.create,
                          Mock(),
                          self.cluster_name,
                          self.datastore,
                          self.datastore_version,
                          self.instances_w_volumes,
                          {}, None, None)

    @patch.object(clients, 'create_nova_client')
    @patch.object(redis_api, 'CONF')
    def test_create_volume_no_specified(self, mock_conf, mock_client):
        mock_conf.get = Mock(
            return_value=FakeOptGroup(volume_support=True))
        self.assertRaises(exception.ClusterVolumeSizeRequired,
                          Cluster.create,
                          Mock(),
                          self.cluster_name,
                          self.datastore,
                          self.datastore_version,
                          self.instances_no_volumes,
                          {}, None, None)

    @patch.object(clients, 'create_nova_client')
    @patch.object(redis_api, 'CONF')
    def test_create_storage_specified_with_no_volume_support(self,
                                                             mock_conf,
                                                             mock_client):
        mock_conf.get = Mock(
            return_value=FakeOptGroup(volume_support=False))
        mock_client.return_value.flavors = Mock()
        self.assertRaises(exception.VolumeNotSupported,
                          Cluster.create,
                          Mock(),
                          self.cluster_name,
                          self.datastore,
                          self.datastore_version,
                          self.instances_w_volumes,
                          {}, None, None)

    @patch.object(clients, 'create_nova_client')
    @patch.object(redis_api, 'CONF')
    def test_create_storage_not_specified_and_no_ephemeral_flavor(self,
                                                                  mock_conf,
                                                                  mock_client):
        class FakeFlavor(object):
            def __init__(self, flavor_id):
                self.flavor_id = flavor_id

            @property
            def id(self):
                return self.flavor.id

            @property
            def ephemeral(self):
                return 0
        mock_conf.get = Mock(
            return_value=FakeOptGroup(volume_support=False))
        (mock_client.return_value.
         flavors.get.return_value) = FakeFlavor('1234')
        self.assertRaises(exception.LocalStorageNotSpecified,
                          Cluster.create,
                          Mock(),
                          self.cluster_name,
                          self.datastore,
                          self.datastore_version,
                          self.instances_no_volumes,
                          {}, None, None)

    @patch.object(redis_api, 'CONF')
    @patch.object(inst_models.Instance, 'create')
    @patch.object(task_api, 'load')
    @patch.object(QUOTAS, 'check_quotas')
    @patch.object(clients, 'create_nova_client')
    def test_create(self, mock_client, mock_check_quotas, mock_task_api,
                    mock_ins_create, mock_conf):
        mock_conf.get = Mock(
            return_value=FakeOptGroup(volume_support=True))
        mock_client.return_value.flavors = Mock()
        self.cluster.create(Mock(),
                            self.cluster_name,
                            self.datastore,
                            self.datastore_version,
                            self.instances_w_volumes, {}, None, None)
        mock_task_api.return_value.create_cluster.assert_called_with(
            self.dbcreate_mock.return_value.id)
        self.assertEqual(3, mock_ins_create.call_count)

    @patch.object(redis_api, 'CONF')
    @patch.object(inst_models.Instance, 'create')
    @patch.object(task_api, 'load')
    @patch.object(QUOTAS, 'check_quotas')
    @patch.object(clients, 'create_nova_client')
    def test_create_with_ephemeral_flavor(self, mock_client, mock_check_quotas,
                                          mock_task_api, mock_ins_create,
                                          mock_conf):
        class FakeFlavor(object):
            def __init__(self, flavor_id):
                self.flavor_id = flavor_id

            @property
            def id(self):
                return self.flavor.id

            @property
            def ephemeral(self):
                return 1
        mock_conf.get = Mock(
            return_value=FakeOptGroup(volume_support=False))
        (mock_client.return_value.
         flavors.get.return_value) = FakeFlavor('1234')
        self.cluster.create(Mock(),
                            self.cluster_name,
                            self.datastore,
                            self.datastore_version,
                            self.instances_no_volumes, {}, None, None)
        mock_task_api.return_value.create_cluster.assert_called_with(
            self.dbcreate_mock.return_value.id)
        self.assertEqual(3, mock_ins_create.call_count)

    @patch.object(DBCluster, 'update')
    @patch.object(redis_api, 'CONF')
    @patch.object(inst_models.Instance, 'create')
    @patch.object(task_api, 'load')
    @patch.object(QUOTAS, 'check_quotas')
    @patch.object(clients, 'create_nova_client')
    def test_grow(self, mock_client, mock_check_quotas, mock_task_api,
                  mock_ins_create, mock_conf, mock_update):
        mock_conf.get = Mock(
            return_value=FakeOptGroup(volume_support=True))
        mock_client.return_value.flavors = Mock()
        self.cluster.grow(self.instances_w_volumes)
        mock_task_api.return_value.grow_cluster.assert_called_with(
            self.dbcreate_mock.return_value.id,
            [mock_ins_create.return_value.id] * 3)
        self.assertEqual(3, mock_ins_create.call_count)

    @patch.object(DBInstance, 'find_all')
    @patch.object(Cluster, 'get_guest')
    @patch.object(DBCluster, 'update')
    @patch.object(inst_models.Instance, 'load')
    @patch.object(inst_models.Instance, 'delete')
    def test_shrink(self,
                    mock_ins_delete, mock_ins_load, mock_update,
                    mock_guest, mock_find_all):
        mock_find_all.return_value.all.return_value = [
            DBInstance(InstanceTasks.NONE, id="1", name="member1",
                       compute_instance_id="compute-1",
                       task_id=InstanceTasks.NONE._code,
                       task_description=InstanceTasks.NONE._db_text,
                       volume_id="volume-1",
                       datastore_version_id="1",
                       cluster_id=self.cluster_id,
                       type="member")]
        self.cluster.shrink(['id1'])
        self.assertEqual(1, mock_ins_delete.call_count)

    @patch('trove.cluster.models.LOG')
    def test_delete_bad_task_status(self, mock_logging):
        self.cluster.db_info.task_status = ClusterTasks.BUILDING_INITIAL
        self.assertRaises(exception.UnprocessableEntity,
                          self.cluster.delete)

    @patch.object(task_api.API, 'delete_cluster')
    @patch.object(Cluster, 'update_db')
    @patch.object(inst_models.DBInstance, 'find_all')
    def test_delete_task_status_none(self,
                                     mock_find_all,
                                     mock_update_db,
                                     mock_delete_cluster):
        self.cluster.db_info.task_status = ClusterTasks.NONE
        self.cluster.delete()
        mock_update_db.assert_called_with(task_status=ClusterTasks.DELETING)

    @patch.object(task_api.API, 'delete_cluster')
    @patch.object(Cluster, 'update_db')
    @patch.object(inst_models.DBInstance, 'find_all')
    def test_delete_task_status_deleting(self,
                                         mock_find_all,
                                         mock_update_db,
                                         mock_delete_cluster):
        self.cluster.db_info.task_status = ClusterTasks.DELETING
        self.cluster.delete()
        mock_update_db.assert_called_with(task_status=ClusterTasks.DELETING)



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc
import re

import sqlalchemy
import urllib

from oslo_log import log as logging
from oslo_utils import encodeutils
from sqlalchemy import event
from sqlalchemy import exc
from sqlalchemy.sql.expression import text

from trove.common import cfg
from trove.common import exception
from trove.common import utils
from trove.common.configurations import MySQLConfParser
from trove.common.db.mysql import models
from trove.common.i18n import _
from trove.guestagent.common import guestagent_utils
from trove.guestagent.common import operating_system
from trove.guestagent.common import sql_query
from trove.guestagent.common.configuration import ConfigurationManager
from trove.guestagent.common.configuration import ImportOverrideStrategy
from trove.guestagent.datastore import service
from trove.guestagent.utils import docker as docker_util
from trove.guestagent.utils import mysql as mysql_util
from trove.instance import service_status

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
ADMIN_USER_NAME = "os_admin"
CONNECTION_STR_FORMAT = ("mysql+pymysql://%s:%s@localhost/?"
                         "unix_socket=/var/run/mysqld/mysqld.sock")
ENGINE = None
INCLUDE_MARKER_OPERATORS = {
    True: ">=",
    False: ">"
}
MYSQL_CONFIG = "/etc/mysql/my.cnf"
CNF_EXT = 'cnf'
CNF_INCLUDE_DIR = '/etc/mysql/conf.d'
CNF_MASTER = 'master-replication'
CNF_SLAVE = 'slave-replication'

BACKUP_LOG = re.compile(r'.*Backup successfully, checksum: (?P<checksum>.*), '
                        r'location: (?P<location>.*)')


class BaseMySqlAppStatus(service.BaseDbStatus):

    def __init__(self, docker_client):
        super(BaseMySqlAppStatus, self).__init__(docker_client)

    def get_actual_db_status(self):
        """Check database service status."""
        status = docker_util.get_container_status(self.docker_client)
        if status == "running":
            root_pass = service.BaseDbApp.get_auth_password(file="root.cnf")
            cmd = 'mysql -uroot -p%s -e "select 1;"' % root_pass
            try:
                docker_util.run_command(self.docker_client, cmd)
                return service_status.ServiceStatuses.HEALTHY
            except Exception as exc:
                LOG.warning('Failed to run docker command, error: %s',
                            str(exc))
                container_log = docker_util.get_container_logs(
                    self.docker_client, tail='all')
                LOG.debug('container log: \n%s', '\n'.join(container_log))
                return service_status.ServiceStatuses.RUNNING
        elif status == "not running":
            return service_status.ServiceStatuses.SHUTDOWN
        elif status == "restarting":
            return service_status.ServiceStatuses.SHUTDOWN
        elif status == "paused":
            return service_status.ServiceStatuses.PAUSED
        elif status == "exited":
            return service_status.ServiceStatuses.SHUTDOWN
        elif status == "dead":
            return service_status.ServiceStatuses.CRASHED
        else:
            return service_status.ServiceStatuses.UNKNOWN


class BaseMySqlAdmin(object, metaclass=abc.ABCMeta):
    """Handles administrative tasks on the MySQL database."""

    def __init__(self, mysql_root_access, mysql_app):
        self.mysql_root_access = mysql_root_access
        self.mysql_app = mysql_app

    def _associate_dbs(self, user):
        """Internal. Given a MySQLUser, populate its databases attribute."""
        LOG.debug("Associating dbs to user %(name)s at %(host)s.",
                  {'name': user.name, 'host': user.host})
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            q = sql_query.Query()
            q.columns = ["grantee", "table_schema"]
            q.tables = ["information_schema.SCHEMA_PRIVILEGES"]
            q.group = ["grantee", "table_schema"]
            q.where = ["privilege_type != 'USAGE'"]
            t = text(str(q))
            db_result = client.execute(t)
            for db in db_result:
                LOG.debug("\t db: %s.", db)
                if db['grantee'] == "'%s'@'%s'" % (user.name, user.host):
                    user.databases = db['table_schema']

    def change_passwords(self, users):
        """Change the passwords of one or more existing users."""
        LOG.debug("Changing the password of some users.")
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            for item in users:
                LOG.debug("Changing password for user %s.", item)
                user_dict = {'_name': item['name'],
                             '_host': item['host'],
                             '_password': item['password']}
                user = models.MySQLUser.deserialize(user_dict)
                uu = sql_query.SetPassword(user.name, host=user.host,
                                           new_password=user.password,
                                           ds=CONF.datastore_manager,
                                           ds_version=CONF.datastore_version)
                t = text(str(uu))
                client.execute(t)

    def update_attributes(self, username, hostname, user_attrs):
        """Change the attributes of an existing user."""
        LOG.debug("Changing user attributes for user %s.", username)
        user = self._get_user(username, hostname)

        new_name = user_attrs.get('name')
        new_host = user_attrs.get('host')
        new_password = user_attrs.get('password')

        if new_name or new_host or new_password:
            with mysql_util.SqlClient(
                self.mysql_app.get_engine(), use_flush=True) as client:
                if new_password is not None:
                    uu = sql_query.SetPassword(
                        user.name, host=user.host,
                        new_password=new_password,
                        ds=CONF.datastore_manager,
                        ds_version=CONF.datastore_version)
                    t = text(str(uu))
                    client.execute(t)

                if new_name or new_host:
                    uu = sql_query.RenameUser(user.name, host=user.host,
                                              new_user=new_name,
                                              new_host=new_host)
                    t = text(str(uu))
                    client.execute(t)

    def create_databases(self, databases):
        """Create the list of specified databases."""
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            for item in databases:
                mydb = models.MySQLSchema.deserialize(item)
                mydb.check_create()
                cd = sql_query.CreateDatabase(mydb.name,
                                              mydb.character_set,
                                              mydb.collate)
                t = text(str(cd))
                LOG.debug('Creating database, command: %s', str(cd))
                client.execute(t)

    def create_users(self, users):
        """Create users and grant them privileges for the
           specified databases.
        """
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            for item in users:
                user = models.MySQLUser.deserialize(item)
                user.check_create()

                cu = sql_query.CreateUser(user.name, host=user.host,
                                          clear=user.password)
                t = text(str(cu))
                client.execute(t, **cu.keyArgs)

                for database in user.databases:
                    mydb = models.MySQLSchema.deserialize(database)
                    g = sql_query.Grant(permissions='ALL', database=mydb.name,
                                        user=user.name, host=user.host)
                    t = text(str(g))
                    LOG.debug('Creating user, command: %s', str(g))
                    client.execute(t)

    def delete_database(self, database):
        """Delete the specified database."""
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            mydb = models.MySQLSchema.deserialize(database)
            mydb.check_delete()
            dd = sql_query.DropDatabase(mydb.name)
            t = text(str(dd))
            client.execute(t)

    def delete_user(self, user):
        """Delete the specified user."""
        mysql_user = models.MySQLUser.deserialize(user)
        mysql_user.check_delete()
        self.delete_user_by_name(mysql_user.name, mysql_user.host)

    def delete_user_by_name(self, name, host='%'):
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            du = sql_query.DropUser(name, host=host)
            t = text(str(du))
            LOG.debug("delete_user_by_name: %s", t)
            client.execute(t)

    def get_user(self, username, hostname):
        user = self._get_user(username, hostname)
        if not user:
            return None
        return user.serialize()

    def _get_user(self, username, hostname):
        """Return a single user matching the criteria."""
        user = None
        try:
            # Could possibly throw a ValueError here.
            user = models.MySQLUser(name=username)
            user.check_reserved()
        except ValueError as ve:
            LOG.exception("Error Getting user information")
            err_msg = encodeutils.exception_to_unicode(ve)
            raise exception.BadRequest(_("Username %(user)s is not valid"
                                         ": %(reason)s") %
                                       {'user': username, 'reason': err_msg}
                                       )
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            q = sql_query.Query()
            q.columns = ['User', 'Host']
            q.tables = ['mysql.user']
            q.where = ["Host != 'localhost'",
                       "User = '%s'" % username,
                       "Host = '%s'" % hostname]
            q.order = ['User', 'Host']
            t = text(str(q))
            result = client.execute(t).fetchall()
            LOG.debug("Getting user information %s.", result)
            if len(result) != 1:
                return None
            found_user = result[0]
            user.host = found_user['Host']
            self._associate_dbs(user)
            return user

    def grant_access(self, username, hostname, databases):
        """Grant a user permission to use a given database."""
        user = self._get_user(username, hostname)
        mydb = None  # cache the model as we just want name validation
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            for database in databases:
                try:
                    if mydb:
                        mydb.name = database
                    else:
                        mydb = models.MySQLSchema(name=database)
                        mydb.check_reserved()
                except ValueError:
                    LOG.exception("Error granting access")
                    raise exception.BadRequest(_(
                        "Grant access to %s is not allowed") % database)

                g = sql_query.Grant(permissions='ALL', database=mydb.name,
                                    user=user.name, host=user.host,
                                    hashed=user.password)
                t = text(str(g))
                client.execute(t)

    def is_root_enabled(self):
        """Return True if root access is enabled; False otherwise."""
        return self.mysql_root_access.is_root_enabled()

    def enable_root(self, root_password=None):
        """Enable the root user global access and/or
           reset the root password.
        """
        return self.mysql_root_access.enable_root(root_password)

    def disable_root(self):
        """Disable the root user global access
        """
        return self.mysql_root_access.disable_root()

    def list_databases(self, limit=None, marker=None, include_marker=False):
        """List databases on this mysql instance."""
        LOG.info("Listing Databases")
        ignored_database_names = "'%s'" % "', '".join(cfg.get_ignored_dbs())
        LOG.debug("The following database names are on ignore list and will "
                  "be omitted from the listing: %s", ignored_database_names)
        databases = []
        with mysql_util.SqlClient(self.mysql_app.get_engine()) as client:
            # If you have an external volume mounted at /var/lib/mysql
            # the lost+found directory will show up in mysql as a database
            # which will create errors if you try to do any database ops
            # on it.  So we remove it here if it exists.
            q = sql_query.Query()
            q.columns = [
                'schema_name as name',
                'default_character_set_name as charset',
                'default_collation_name as collation',
            ]
            q.tables = ['information_schema.schemata']
            q.where = ["schema_name NOT IN (" + ignored_database_names + ")"]
            q.order = ['schema_name ASC']
            if limit:
                q.limit = limit + 1
            if marker:
                q.where.append("schema_name %s '%s'" %
                               (INCLUDE_MARKER_OPERATORS[include_marker],
                                marker))
            t = text(str(q))
            database_names = client.execute(t)
            next_marker = None
            for count, database in enumerate(database_names):
                if limit is not None and count >= limit:
                    break
                mysql_db = models.MySQLSchema(name=database[0],
                                              character_set=database[1],
                                              collate=database[2])
                next_marker = mysql_db.name
                databases.append(mysql_db.serialize())

        LOG.info("databases = %s", str(databases))
        if limit is not None and database_names.rowcount <= limit:
            next_marker = None
        return databases, next_marker

    def list_users(self, limit=None, marker=None, include_marker=False):
        """List users that have access to the database."""
        '''
        SELECT
            User,
            Host,
            Marker
        FROM
            (SELECT
                User,
                Host,
                CONCAT(User, '@', Host) as Marker
            FROM mysql.user
            ORDER BY 1, 2) as innerquery
        WHERE
            Marker > :marker
        ORDER BY
            Marker
        LIMIT :limit;
        '''
        LOG.info("Listing Users")
        ignored_user_names = "'%s'" % "', '".join(cfg.get_ignored_users())
        LOG.debug("The following user names are on ignore list and will "
                  "be omitted from the listing: %s", ignored_user_names)
        users = []
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            iq = sql_query.Query()  # Inner query.
            iq.columns = ['User', 'Host', "CONCAT(User, '@', Host) as Marker"]
            iq.tables = ['mysql.user']
            iq.order = ['User', 'Host']
            innerquery = str(iq).rstrip(';')

            oq = sql_query.Query()  # Outer query.
            oq.columns = ['User', 'Host', 'Marker']
            oq.tables = ['(%s) as innerquery' % innerquery]
            oq.where = [
                "Host != 'localhost'",
                "User NOT IN (" + ignored_user_names + ")"]
            oq.order = ['Marker']
            if marker:
                oq.where.append("Marker %s '%s'" %
                                (INCLUDE_MARKER_OPERATORS[include_marker],
                                 marker))
            if limit:
                oq.limit = limit + 1
            t = text(str(oq))
            result = client.execute(t)
            next_marker = None
            for count, row in enumerate(result):
                if limit is not None and count >= limit:
                    break
                LOG.debug("user = %s", str(row))
                mysql_user = models.MySQLUser(name=row['User'],
                                              host=row['Host'])
                mysql_user.check_reserved()
                self._associate_dbs(mysql_user)
                next_marker = row['Marker']
                users.append(mysql_user.serialize())
        if limit is not None and result.rowcount <= limit:
            next_marker = None
        LOG.info("users = %s", str(users))

        return users, next_marker

    def revoke_access(self, username, hostname, database):
        """Revoke a user's permission to use a given database."""
        user = self._get_user(username, hostname)
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            r = sql_query.Revoke(database=database,
                                 user=user.name,
                                 host=user.host)
            t = text(str(r))
            client.execute(t)

    def list_access(self, username, hostname):
        """Show all the databases to which the user has more than
           USAGE granted.
        """
        user = self._get_user(username, hostname)
        return user.databases


class BaseMySqlApp(service.BaseDbApp):
    _configuration_manager = None

    @property
    def configuration_manager(self):
        if self._configuration_manager:
            return self._configuration_manager

        self._configuration_manager = ConfigurationManager(
            MYSQL_CONFIG, CONF.database_service_uid, CONF.database_service_uid,
            service.BaseDbApp.CFG_CODEC, requires_root=True,
            override_strategy=ImportOverrideStrategy(CNF_INCLUDE_DIR, CNF_EXT)
        )
        return self._configuration_manager

    def get_engine(self):
        """Create the default engine with the updated admin user.

        If admin user not created yet, use root instead.
        """
        global ENGINE
        if ENGINE:
            return ENGINE

        try:
            user = ADMIN_USER_NAME
            password = self.get_auth_password()
        except exception.UnprocessableEntity:
            # os_admin user not created yet
            user = 'root'
            password = self.get_auth_password(file="root.cnf")

        ENGINE = sqlalchemy.create_engine(
            CONNECTION_STR_FORMAT % (user,
                                     urllib.parse.quote(password.strip())),
            pool_recycle=120, echo=CONF.sql_query_logging
        )
        event.listen(ENGINE, 'checkout', mysql_util.connection_checkout)

        return ENGINE

    def execute_sql(self, sql_statement, use_flush=False):
        LOG.debug("Executing SQL: %s", sql_statement)
        with mysql_util.SqlClient(
            self.get_engine(), use_flush=use_flush) as client:
            return client.execute(sql_statement)

    def get_data_dir(self):
        return self.configuration_manager.get_value(
            'datadir', section=MySQLConfParser.SERVER_CONF_SECTION)

    def set_data_dir(self, value):
        self.configuration_manager.apply_system_override(
            {MySQLConfParser.SERVER_CONF_SECTION: {'datadir': value}})

    def _create_admin_user(self, client, password):
        """
        Create a os_admin user with a random password
        with all privileges similar to the root user.
        """
        LOG.info("Creating Trove admin user '%s'.", ADMIN_USER_NAME)
        host = "localhost"
        try:
            cu = sql_query.CreateUser(ADMIN_USER_NAME, host=host,
                                      clear=password)
            t = text(str(cu))
            client.execute(t, **cu.keyArgs)
        except (exc.OperationalError, exc.InternalError) as err:
            # Ignore, user is already created, just reset the password
            # (user will already exist in a restore from backup)
            LOG.debug(err)
            uu = sql_query.SetPassword(
                ADMIN_USER_NAME, host=host, new_password=password,
                ds=CONF.datastore_manager, ds_version=CONF.datastore_version
            )
            t = text(str(uu))
            client.execute(t)

        g = sql_query.Grant(permissions='ALL', user=ADMIN_USER_NAME,
                            host=host, grant_option=True)
        t = text(str(g))
        client.execute(t)
        LOG.info("Trove admin user '%s' created.", ADMIN_USER_NAME)

    def secure(self):
        LOG.info("Securing MySQL now.")

        root_pass = self.get_auth_password(file="root.cnf")
        admin_password = utils.generate_random_password()

        engine = sqlalchemy.create_engine(
            CONNECTION_STR_FORMAT % ('root', root_pass), echo=True)
        with mysql_util.SqlClient(engine) as client:
            self._create_admin_user(client, admin_password)

        engine = sqlalchemy.create_engine(
            CONNECTION_STR_FORMAT % (ADMIN_USER_NAME,
                                     urllib.parse.quote(admin_password)),
            echo=True)
        with mysql_util.SqlClient(engine, use_flush=True) as client:
            self._remove_anonymous_user(client)

        self.save_password(ADMIN_USER_NAME, admin_password)
        LOG.info("MySQL secure complete.")

    def secure_root(self):
        with mysql_util.SqlClient(self.get_engine(), use_flush=True) as client:
            self._remove_remote_root_access(client)

    def _remove_anonymous_user(self, client):
        LOG.debug("Removing anonymous user.")
        t = text(sql_query.REMOVE_ANON)
        client.execute(t)
        LOG.debug("Anonymous user removed.")

    def _remove_remote_root_access(self, client):
        LOG.debug("Removing remote root access.")
        t = text(sql_query.REMOVE_ROOT)
        client.execute(t)
        LOG.debug("Root remote access removed.")

    def update_overrides(self, overrides):
        if overrides:
            self.configuration_manager.apply_user_override(
                {MySQLConfParser.SERVER_CONF_SECTION: overrides})

    def apply_overrides(self, overrides):
        with mysql_util.SqlClient(self.get_engine()) as client:
            for k, v in overrides.items():
                byte_value = guestagent_utils.to_bytes(v)
                q = sql_query.SetServerVariable(key=k, value=byte_value)
                t = text(str(q))
                try:
                    client.execute(t)
                except exc.OperationalError:
                    output = {'key': k, 'value': byte_value}
                    LOG.error("Unable to set %(key)s with value %(value)s.",
                              output)

    def start_db(self, update_db=False, ds_version=None, command=None,
                 extra_volumes=None):
        """Start and wait for database service."""
        docker_image = CONF.get(CONF.datastore_manager).docker_image
        image = (f'{docker_image}:latest' if not ds_version else
                 f'{docker_image}:{ds_version}')
        command = command if command else ''

        try:
            root_pass = self.get_auth_password(file="root.cnf")
        except exception.UnprocessableEntity:
            root_pass = utils.generate_random_password()

        # Get uid and gid
        user = "%s:%s" % (CONF.database_service_uid, CONF.database_service_uid)

        # Create folders for mysql on localhost
        for folder in ['/etc/mysql', '/var/run/mysqld']:
            operating_system.ensure_directory(
                folder, user=CONF.database_service_uid,
                group=CONF.database_service_uid, force=True,
                as_root=True)

        volumes = {
            "/etc/mysql": {"bind": "/etc/mysql", "mode": "rw"},
            "/var/run/mysqld": {"bind": "/var/run/mysqld",
                                "mode": "rw"},
            "/var/lib/mysql": {"bind": "/var/lib/mysql", "mode": "rw"},
        }
        if extra_volumes:
            volumes.update(extra_volumes)

        # Expose ports
        ports = {}
        tcp_ports = cfg.get_configuration_property('tcp_ports')
        for port_range in tcp_ports:
            for port in port_range:
                ports[f'{port}/tcp'] = port

        try:
            docker_util.start_container(
                self.docker_client,
                image,
                volumes=volumes,
                network_mode="bridge",
                ports=ports,
                user=user,
                environment={
                    "MYSQL_ROOT_PASSWORD": root_pass,
                    "MYSQL_INITDB_SKIP_TZINFO": 1,
                },
                command=command
            )

            # Save root password
            LOG.debug("Saving root credentials to local host.")
            self.save_password('root', root_pass)
        except Exception:
            LOG.exception("Failed to start mysql")
            raise exception.TroveError(_("Failed to start mysql"))

        if not self.status.wait_for_status(
            service_status.ServiceStatuses.HEALTHY,
            CONF.state_change_wait_time, update_db
        ):
            raise exception.TroveError(_("Failed to start mysql"))

    def wipe_ib_logfiles(self):
        """Destroys the iblogfiles.

        If for some reason the selected log size in the conf changes from the
        current size of the files MySQL will fail to start, so we delete the
        files to be safe.
        """
        for index in range(2):
            try:
                # On restarts, sometimes these are wiped. So it can be a race
                # to have MySQL start up before it's restarted and these have
                # to be deleted. That's why its ok if they aren't found and
                # that is why we use the "force" option to "remove".
                operating_system.remove("%s/ib_logfile%d"
                                        % (self.get_data_dir(), index),
                                        force=True, as_root=True)
            except exception.ProcessExecutionError:
                LOG.exception("Could not delete logfile.")
                raise

    def restart(self):
        LOG.info("Restarting mysql")

        # Ensure folders permission for database.
        for folder in ['/etc/mysql', '/var/run/mysqld']:
            operating_system.ensure_directory(
                folder, user=CONF.database_service_uid,
                group=CONF.database_service_uid, force=True,
                as_root=True)

        try:
            docker_util.restart_container(self.docker_client)
        except Exception:
            LOG.exception("Failed to restart mysql")
            raise exception.TroveError("Failed to restart mysql")

        if not self.status.wait_for_status(
            service_status.ServiceStatuses.HEALTHY,
            CONF.state_change_wait_time, update_db=True
        ):
            raise exception.TroveError("Failed to start mysql")

        LOG.info("Finished restarting mysql")

    def restore_backup(self, context, backup_info, restore_location):
        backup_id = backup_info['id']
        storage_driver = CONF.storage_strategy
        backup_driver = self.get_backup_strategy()
        user_token = context.auth_token
        auth_url = CONF.service_credentials.auth_url
        user_tenant = context.project_id
        image = self.get_backup_image()
        name = 'db_restore'
        volumes = {'/var/lib/mysql': {'bind': '/var/lib/mysql', 'mode': 'rw'}}

        command = (
            f'/usr/bin/python3 main.py --nobackup '
            f'--storage-driver={storage_driver} --driver={backup_driver} '
            f'--os-token={user_token} --os-auth-url={auth_url} '
            f'--os-tenant-id={user_tenant} '
            f'--restore-from={backup_info["location"]} '
            f'--restore-checksum={backup_info["checksum"]}'
        )
        if CONF.backup_aes_cbc_key:
            command = (f"{command} "
                       f"--backup-encryption-key={CONF.backup_aes_cbc_key}")

        LOG.debug('Stop the database and clean up the data before restore '
                  'from %s', backup_id)
        self.stop_db()
        operating_system.chmod(restore_location,
                               operating_system.FileMode.SET_FULL,
                               as_root=True)
        utils.clean_out(restore_location)

        # Start to run restore inside a separate docker container
        LOG.info('Starting to restore backup %s, command: %s', backup_id,
                 command)
        output, ret = docker_util.run_container(
            self.docker_client, image, name,
            volumes=volumes, command=command)
        result = output[-1]
        if not ret:
            msg = f'Failed to run restore container, error: {result}'
            LOG.error(msg)
            raise Exception(msg)

        LOG.debug('Deleting ib_logfile files after restore from backup %s',
                  backup_id)
        operating_system.chown(restore_location, CONF.database_service_uid,
                               CONF.database_service_uid, force=True,
                               as_root=True)
        self.wipe_ib_logfiles()

    def exists_replication_source_overrides(self):
        return self.configuration_manager.has_system_override(CNF_MASTER)

    def write_replication_source_overrides(self, overrideValues):
        self.configuration_manager.apply_system_override(overrideValues,
                                                         CNF_MASTER)

    def write_replication_replica_overrides(self, overrideValues):
        self.configuration_manager.apply_system_override(overrideValues,
                                                         CNF_SLAVE)

    def remove_replication_source_overrides(self):
        self.configuration_manager.remove_system_override(CNF_MASTER)

    def remove_replication_replica_overrides(self):
        self.configuration_manager.remove_system_override(CNF_SLAVE)

    def grant_replication_privilege(self, replication_user):
        LOG.info("Granting replication slave privilege for %s",
                 replication_user['name'])

        with mysql_util.SqlClient(self.get_engine(), use_flush=True) as client:
            g = sql_query.Grant(permissions=['REPLICATION SLAVE'],
                                user=replication_user['name'],
                                clear=replication_user['password'])

            t = text(str(g))
            client.execute(t)

    def get_port(self):
        with mysql_util.SqlClient(self.get_engine()) as client:
            result = client.execute('SELECT @@port').first()
            return result[0]

    def wait_for_slave_status(self, status, client, max_time):
        def verify_slave_status():
            ret = client.execute(
                "SELECT SERVICE_STATE FROM "
                "performance_schema.replication_connection_status").first()
            if not ret:
                actual_status = 'OFF'
            else:
                actual_status = ret[0]
            return actual_status.upper() == status.upper()

        LOG.debug("Waiting for slave status %s with timeout %s",
                  status, max_time)
        try:
            utils.poll_until(verify_slave_status, sleep_time=3,
                             time_out=max_time)
            LOG.info("Replication status: %s.", status)
        except exception.PollTimeOut:
            raise RuntimeError(
                _("Replication is not %(status)s after %(max)d seconds.") % {
                    'status': status.lower(), 'max': max_time})

    def start_slave(self):
        LOG.info("Starting slave replication.")
        with mysql_util.SqlClient(self.get_engine()) as client:
            client.execute('START SLAVE')
            self.wait_for_slave_status("ON", client, 180)

    def stop_slave(self, for_failover):
        LOG.info("Stopping slave replication.")

        replication_user = None
        with mysql_util.SqlClient(self.get_engine()) as client:
            result = client.execute('SHOW SLAVE STATUS')
            replication_user = result.first()['Master_User']
            client.execute('STOP SLAVE')
            client.execute('RESET SLAVE ALL')
            self.wait_for_slave_status('OFF', client, 180)
            if not for_failover:
                client.execute('DROP USER IF EXISTS ' + replication_user)

        return {
            'replication_user': replication_user
        }

    def stop_master(self):
        LOG.info("Stopping replication master.")
        with mysql_util.SqlClient(self.get_engine()) as client:
            client.execute('RESET MASTER')

    def make_read_only(self, read_only):
        with mysql_util.SqlClient(self.get_engine()) as client:
            q = "set global read_only = %s" % read_only
            client.execute(text(str(q)))

    def upgrade(self, upgrade_info):
        """Upgrade the database."""
        new_version = upgrade_info.get('datastore_version')
        if new_version == CONF.datastore_version:
            return

        LOG.info('Stopping db container for upgrade')
        self.stop_db()

        LOG.info('Deleting db container for upgrade')
        docker_util.remove_container(self.docker_client)

        LOG.info('Remove unused images before starting new db container')
        docker_util.prune_images(self.docker_client)

        LOG.info('Starting new db container with version %s for upgrade',
                 new_version)
        self.start_db(update_db=True, ds_version=new_version)


class BaseMySqlRootAccess(object):

    def __init__(self, mysql_app):
        self.mysql_app = mysql_app

    def is_root_enabled(self):
        """Return True if root access is enabled; False otherwise."""
        with mysql_util.SqlClient(self.mysql_app.get_engine()) as client:
            t = text(sql_query.ROOT_ENABLED)
            result = client.execute(t)
            LOG.debug("Found %s with remote root access.", result.rowcount)
            return result.rowcount != 0

    def enable_root(self, root_password=None):
        """Enable the root user global access and/or
           reset the root password.
        """
        user = models.MySQLUser.root(password=root_password)
        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            try:
                cu = sql_query.CreateUser(user.name, host=user.host)
                t = text(str(cu))
                client.execute(t, **cu.keyArgs)
            except (exc.OperationalError, exc.InternalError) as err:
                # Ignore, user is already created, just reset the password
                # TODO(rnirmal): More fine grained error checking later on
                LOG.debug(err)

        with mysql_util.SqlClient(
            self.mysql_app.get_engine(), use_flush=True) as client:
            uu = sql_query.SetPassword(
                user.name, host=user.host, new_password=user.password,
                ds=CONF.datastore_manager, ds_version=CONF.datastore_version
            )
            t = text(str(uu))
            client.execute(t)

            LOG.debug("CONF.root_grant: %(grant)s CONF.root_grant_option: "
                      "%(grant_option)s.",
                      {'grant': CONF.root_grant,
                       'grant_option': CONF.root_grant_option})

            g = sql_query.Grant(permissions=CONF.root_grant,
                                user=user.name,
                                host=user.host,
                                grant_option=CONF.root_grant_option)

            t = text(str(g))
            client.execute(t)
            return user.serialize()

    def disable_root(self):
        """Reset the root password to an unknown value.
        """
        self.enable_root(root_password=None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob

from trove.common import auth
from trove.tests.unittests import trove_testtools


class TestAuth(trove_testtools.TestCase):
    def test_unicode_characters_in_headers(self):
        middleware = auth.AuthorizationMiddleware(
            "test_trove",
            [auth.TenantBasedAuth()])
        tenant_id = 'test_tenant_id'
        url = '/%s/instances' % tenant_id
        req = webob.Request.blank(url)

        # test string with chinese characters
        test_str = '\u6d4b\u8bd5'
        req.headers = {
            'X-Tenant-ID': tenant_id,
            'X-Auth-Project-Id': test_str
        }
        # invocation
        middleware.process_request(req)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from time import sleep

from proboscis.asserts import assert_equal
from proboscis.asserts import assert_raises
from proboscis.asserts import assert_true
from proboscis.asserts import fail
from proboscis.decorators import time_out
from proboscis import SkipTest
from proboscis import test
from troveclient.compat import exceptions

from trove.common.utils import generate_uuid
from trove.common.utils import poll_until
from trove import tests
from trove.tests.api.instances import CheckInstance
from trove.tests.api.instances import instance_info
from trove.tests.api.instances import TIMEOUT_INSTANCE_DELETE
from trove.tests.api.instances import TIMEOUT_INSTANCE_RESTORE
from trove.tests.config import CONFIG
from trove.tests.scenario import runners
from trove.tests.scenario.runners.test_runners import SkipKnownBug
from trove.tests.util.server_connection import create_server_connection


class SlaveInstanceTestInfo(object):
    """Stores slave instance information."""
    def __init__(self):
        self.id = None
        self.replicated_db = generate_uuid()


slave_instance = SlaveInstanceTestInfo()
existing_db_on_master = generate_uuid()
backup_count = None


def _get_user_count(server_info):
    cmd = (
        'docker exec -e MYSQL_PWD=$(sudo cat /var/lib/mysql/conf.d/root.cnf | '
        'grep password | awk "{print \$3}") database mysql -uroot -N -e '
        '"select count(*) from mysql.user where user like \\"slave_%\\""'
    )
    server = create_server_connection(server_info.id)

    try:
        stdout = server.execute(cmd)
        return int(stdout.rstrip())
    except Exception as e:
        fail("Failed to execute command: %s, error: %s" % (cmd, str(e)))


def slave_is_running(running=True):
    def check_slave_is_running():
        server = create_server_connection(slave_instance.id)
        cmd = (
            'docker exec -e MYSQL_PWD=$(sudo cat '
            '/var/lib/mysql/conf.d/root.cnf | grep password '
            '| awk "{print \$3}") database mysql -uroot -N -e '
            '"SELECT SERVICE_STATE FROM '
            'performance_schema.replication_connection_status"'
        )

        try:
            stdout = server.execute(cmd)
            stdout = stdout.rstrip()
        except Exception as e:
            fail("Failed to execute command %s, error: %s" %
                 (cmd, str(e)))

        expected = b"ON" if running else b""
        return stdout == expected

    return check_slave_is_running


def backup_count_matches(count):

    def check_backup_count_matches():
        backup = instance_info.dbaas.instances.backups(instance_info.id)
        return count == len(backup)

    return check_backup_count_matches


def instance_is_active(id):
    instance = instance_info.dbaas.instances.get(id)
    if instance.status in CONFIG.running_status:
        return True
    else:
        assert_true(instance.status in ['PROMOTE', 'EJECT', 'BUILD', 'BACKUP'])
        return False


def create_slave():
    result = instance_info.dbaas.instances.create(
        instance_info.name + "_slave",
        nics=instance_info.nics,
        replica_of=instance_info.id)
    assert_equal(200, instance_info.dbaas.last_http_code)
    assert_equal("BUILD", result.status)
    return result.id


def validate_slave(master, slave):
    new_slave = instance_info.dbaas.instances.get(slave.id)
    assert_equal(200, instance_info.dbaas.last_http_code)
    ns_dict = new_slave._info
    CheckInstance(ns_dict).replica_of()
    assert_equal(master.id, ns_dict['replica_of']['id'])


def validate_master(master, slaves):
    new_master = instance_info.dbaas.instances.get(master.id)
    assert_equal(200, instance_info.dbaas.last_http_code)
    nm_dict = new_master._info
    CheckInstance(nm_dict).slaves()
    master_ids = set([replica['id'] for replica in nm_dict['replicas']])
    asserted_ids = set([slave.id for slave in slaves])
    assert_true(asserted_ids.issubset(master_ids))


@test(depends_on_groups=[tests.DBAAS_API_CONFIGURATIONS],
      groups=[tests.DBAAS_API_REPLICATION],
      enabled=CONFIG.swift_enabled)
class CreateReplicationSlave(object):
    @test
    def test_create_db_on_master(self):
        """test_create_db_on_master"""
        databases = [{'name': existing_db_on_master}]
        # Ensure that the auth_token in the dbaas client is not stale
        instance_info.dbaas.authenticate()
        instance_info.dbaas.databases.create(instance_info.id, databases)
        assert_equal(202, instance_info.dbaas.last_http_code)

    @test(runs_after=['test_create_db_on_master'])
    def test_create_slave(self):
        """test_create_slave"""
        global backup_count
        backup_count = len(
            instance_info.dbaas.instances.backups(instance_info.id))
        slave_instance.id = create_slave()


@test(groups=[tests.DBAAS_API_REPLICATION],
      enabled=CONFIG.swift_enabled,
      depends_on_classes=[CreateReplicationSlave])
class WaitForCreateSlaveToFinish(object):
    """Wait until the instance is created and set up as slave."""

    @test
    @time_out(TIMEOUT_INSTANCE_RESTORE)
    def test_slave_created(self):
        """Wait for replica to be created."""
        poll_until(lambda: instance_is_active(slave_instance.id))


@test(enabled=(not CONFIG.fake_mode and CONFIG.swift_enabled),
      depends_on_classes=[WaitForCreateSlaveToFinish],
      groups=[tests.DBAAS_API_REPLICATION])
class VerifySlave(object):

    def db_is_found(self, database_to_find):

        def find_database():
            databases = instance_info.dbaas.databases.list(slave_instance.id)
            return (database_to_find in [d.name for d in databases])

        return find_database

    @test
    @time_out(10 * 60)
    def test_correctly_started_replication(self):
        """test_correctly_started_replication"""
        poll_until(slave_is_running())

    @test(runs_after=[test_correctly_started_replication])
    @time_out(60)
    def test_backup_deleted(self):
        """test_backup_deleted"""
        poll_until(backup_count_matches(backup_count))

    @test(depends_on=[test_correctly_started_replication])
    def test_slave_is_read_only(self):
        """test_slave_is_read_only"""
        cmd = (
            'docker exec -e MYSQL_PWD=$(sudo cat '
            '/var/lib/mysql/conf.d/root.cnf | grep password | '
            'awk "{print \$3}") database mysql -uroot -NBq -e '
            '"select @@read_only"'
        )
        server = create_server_connection(slave_instance.id)

        try:
            stdout = server.execute(cmd)
            stdout = int(stdout.rstrip())
        except Exception as e:
            fail("Failed to execute command %s, error: %s" %
                 (cmd, str(e)))

        assert_equal(stdout, 1)

    @test(depends_on=[test_slave_is_read_only])
    def test_create_db_on_master(self):
        """test_create_db_on_master"""
        databases = [{'name': slave_instance.replicated_db}]
        instance_info.dbaas.databases.create(instance_info.id, databases)
        assert_equal(202, instance_info.dbaas.last_http_code)

    @test(depends_on=[test_create_db_on_master])
    @time_out(5 * 60)
    def test_database_replicated_on_slave(self):
        """test_database_replicated_on_slave"""
        poll_until(self.db_is_found(slave_instance.replicated_db))

    @test(runs_after=[test_database_replicated_on_slave])
    @time_out(5 * 60)
    def test_existing_db_exists_on_slave(self):
        """test_existing_db_exists_on_slave"""
        poll_until(self.db_is_found(existing_db_on_master))

    @test(depends_on=[test_existing_db_exists_on_slave])
    def test_slave_user_exists(self):
        """test_slave_user_exists"""
        assert_equal(_get_user_count(slave_instance), 1)
        assert_equal(_get_user_count(instance_info), 1)


@test(groups=[tests.DBAAS_API_REPLICATION],
      depends_on_classes=[VerifySlave],
      enabled=CONFIG.swift_enabled)
class TestInstanceListing(object):
    """Test replication information in instance listing."""

    @test
    def test_get_slave_instance(self):
        """test_get_slave_instance"""
        validate_slave(instance_info, slave_instance)

    @test
    def test_get_master_instance(self):
        """test_get_master_instance"""
        validate_master(instance_info, [slave_instance])


@test(groups=[tests.DBAAS_API_REPLICATION],
      depends_on_classes=[TestInstanceListing],
      enabled=CONFIG.swift_enabled)
class TestReplicationFailover(object):
    """Test replication failover functionality."""

    @staticmethod
    def promote(master, slave):
        if CONFIG.fake_mode:
            raise SkipTest("promote_replica_source not supported in fake mode")

        instance_info.dbaas.instances.promote_to_replica_source(slave)
        assert_equal(202, instance_info.dbaas.last_http_code)
        poll_until(lambda: instance_is_active(slave.id))
        validate_master(slave, [master])
        validate_slave(slave, master)

    @test
    def test_promote_master(self):
        if CONFIG.fake_mode:
            raise SkipTest("promote_master not supported in fake mode")

        assert_raises(exceptions.BadRequest,
                      instance_info.dbaas.instances.promote_to_replica_source,
                      instance_info.id)

    @test
    def test_eject_slave(self):
        if CONFIG.fake_mode:
            raise SkipTest("eject_replica_source not supported in fake mode")

        assert_raises(exceptions.BadRequest,
                      instance_info.dbaas.instances.eject_replica_source,
                      slave_instance.id)

    @test
    def test_eject_valid_master(self):
        if CONFIG.fake_mode:
            raise SkipTest("eject_replica_source not supported in fake mode")

        # assert_raises(exceptions.BadRequest,
        #               instance_info.dbaas.instances.eject_replica_source,
        #               instance_info.id)
        # Uncomment once BUG_EJECT_VALID_MASTER is fixed
        raise SkipKnownBug(runners.BUG_EJECT_VALID_MASTER)

    @test(depends_on=[test_promote_master, test_eject_slave,
                      test_eject_valid_master])
    def test_promote_to_replica_source(self):
        """test_promote_to_replica_source"""
        TestReplicationFailover.promote(instance_info, slave_instance)

    @test(depends_on=[test_promote_to_replica_source])
    def test_promote_back_to_replica_source(self):
        """test_promote_back_to_replica_source"""
        TestReplicationFailover.promote(slave_instance, instance_info)

    @test(depends_on=[test_promote_back_to_replica_source], enabled=False)
    def add_second_slave(self):
        """add_second_slave"""
        if CONFIG.fake_mode:
            raise SkipTest("three site promote not supported in fake mode")

        self._third_slave = SlaveInstanceTestInfo()
        self._third_slave.id = create_slave()
        poll_until(lambda: instance_is_active(self._third_slave.id))
        poll_until(slave_is_running())
        sleep(15)
        validate_master(instance_info, [slave_instance, self._third_slave])
        validate_slave(instance_info, self._third_slave)

    @test(depends_on=[add_second_slave], enabled=False)
    def test_three_site_promote(self):
        """Promote the second slave"""
        if CONFIG.fake_mode:
            raise SkipTest("three site promote not supported in fake mode")

        TestReplicationFailover.promote(instance_info, self._third_slave)
        validate_master(self._third_slave, [slave_instance, instance_info])
        validate_slave(self._third_slave, instance_info)

    @test(depends_on=[test_three_site_promote], enabled=False)
    def disable_master(self):
        """Stop trove-guestagent on master"""
        if CONFIG.fake_mode:
            raise SkipTest("eject_replica_source not supported in fake mode")

        cmd = "sudo systemctl stop guest-agent.service"
        server = create_server_connection(self._third_slave.id)

        try:
            stdout = server.execute(cmd)
            stdout = int(stdout.rstrip())
        except Exception as e:
            fail("Failed to execute command %s, error: %s" %
                 (cmd, str(e)))

        assert_equal(stdout, 1)

    @test(depends_on=[disable_master], enabled=False)
    def test_eject_replica_master(self):
        if CONFIG.fake_mode:
            raise SkipTest("eject_replica_source not supported in fake mode")

        sleep(70)
        instance_info.dbaas.instances.eject_replica_source(self._third_slave)
        assert_equal(202, instance_info.dbaas.last_http_code)
        poll_until(lambda: instance_is_active(self._third_slave.id))
        validate_master(instance_info, [slave_instance])
        validate_slave(instance_info, slave_instance)


@test(groups=[tests.DBAAS_API_REPLICATION],
      depends_on=[TestReplicationFailover],
      enabled=CONFIG.swift_enabled)
class DetachReplica(object):

    @test
    def delete_before_detach_replica(self):
        assert_raises(exceptions.Forbidden,
                      instance_info.dbaas.instances.delete,
                      instance_info.id)

    @test
    @time_out(5 * 60)
    def test_detach_replica(self):
        """test_detach_replica"""
        if CONFIG.fake_mode:
            raise SkipTest("Detach replica not supported in fake mode")

        instance_info.dbaas.instances.update(slave_instance.id,
                                             detach_replica_source=True)
        assert_equal(202, instance_info.dbaas.last_http_code)

        poll_until(slave_is_running(False))

    @test(depends_on=[test_detach_replica])
    @time_out(5 * 60)
    def test_slave_is_not_read_only(self):
        """test_slave_is_not_read_only"""
        if CONFIG.fake_mode:
            raise SkipTest("Test not_read_only not supported in fake mode")

        # wait until replica is no longer read only
        def check_not_read_only():
            cmd = (
                'docker exec -e MYSQL_PWD=$(sudo cat '
                '/var/lib/mysql/conf.d/root.cnf | grep password | '
                'awk "{print \$3}") database mysql -uroot -NBq -e '
                '"select @@read_only"'
            )
            server = create_server_connection(slave_instance.id)

            try:
                stdout = server.execute(cmd)
                stdout = int(stdout)
            except Exception:
                return False

            return stdout == 0

        poll_until(check_not_read_only)


@test(groups=[tests.DBAAS_API_REPLICATION],
      depends_on=[DetachReplica],
      enabled=CONFIG.swift_enabled)
class DeleteSlaveInstance(object):

    @test
    @time_out(TIMEOUT_INSTANCE_DELETE)
    def test_delete_slave_instance(self):
        """test_delete_slave_instance"""
        instance_info.dbaas.instances.delete(slave_instance.id)
        assert_equal(202, instance_info.dbaas.last_http_code)

        def instance_is_gone():
            try:
                instance_info.dbaas.instances.get(slave_instance.id)
                return False
            except exceptions.NotFound:
                return True

        poll_until(instance_is_gone)
        assert_raises(exceptions.NotFound, instance_info.dbaas.instances.get,
                      slave_instance.id)



# Copyright 2016 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.instance_force_delete_group"


class InstanceForceDeleteRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'instance_force_delete_runners'
    _runner_cls = 'InstanceForceDeleteRunner'


@test(depends_on_groups=[groups.INST_ERROR_DELETE_WAIT],
      groups=[GROUP, groups.INST_FORCE_DELETE])
class InstanceForceDeleteGroup(TestGroup):
    """Test Instance Force Delete functionality."""

    def __init__(self):
        super(InstanceForceDeleteGroup, self).__init__(
            InstanceForceDeleteRunnerFactory.instance())

    @test
    def create_build_instance(self):
        """Create an instance in BUILD state."""
        self.test_runner.run_create_build_instance()

    @test(depends_on=['create_build_instance'])
    def delete_build_instance(self):
        """Make sure the instance in BUILD state deletes."""
        self.test_runner.run_delete_build_instance()


@test(depends_on_classes=[InstanceForceDeleteGroup],
      groups=[GROUP, groups.INST_FORCE_DELETE_WAIT])
class InstanceForceDeleteWaitGroup(TestGroup):
    """Make sure the Force Delete instance goes away."""

    def __init__(self):
        super(InstanceForceDeleteWaitGroup, self).__init__(
            InstanceForceDeleteRunnerFactory.instance())

    @test
    def wait_for_force_delete(self):
        """Wait for the Force Delete instance to be gone."""
        self.test_runner.run_wait_for_force_delete()
